/******************************************************************************/
/*                                                                            */
/*  <<< Copyright >>>                                                         */
/*                                                                            */
/*  The "xmemo" software is copyright 1999 by Takao Abe.                      */
/*  All rights reserved.                                                      */
/*                                                                            */
/*  The "xmemo" software refers to the source code, the binary code,          */
/*  the documents, and all derivatives based on this software under           */
/*  the copyright law, that is a software containing a program or a           */
/*  portion of a program, regardless of modifications and/or translations.    */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/*  <<< Licence and redistribution >>>                                        */
/*                                                                            */
/*  The source code and the binary code are free software.                    */
/*  The use and redistribution of source code and binary code are permitted   */
/*  provided that the following conditions are met:                           */
/*  You may use the software for private use and commercial use,              */
/*  and may modify your copy or copies of the software or any portion of it.  */
/*                                                                            */
/*  THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESSED OR IMPLIED            */
/*  WARRANTIES OF ANY KIND, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED        */
/*  WARRANTIES OF MERCHANTABLILITY AND FITNESS FOR A PARTICULAR PURPOSE.      */
/*  IN NO EVENT SHALL THE AUTHOR TAKAO ABE BE LIABLE FOR ANY DIRECT,          */
/*  INDIRECT, GENERAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES           */
/*  ( INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR       */
/*  SERVICES; LOSS OF USE, DATA OR PROFITS; OR BUSINESS INTERRUPTION )        */
/*  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,       */
/*  STRICT LIABILITY, OR TORT ( INCLUDING NEGLIGENCE OR OTHERWISE )           */
/*  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED       */
/*  OF THE POSSIBILITY OF SUCH DAMAGE.                                        */
/*                                                                            */
/*  Redistribution and use in source form or binary form, with or without     */
/*  modification, the whole program or a portion of it, are permitted         */
/*  provided that one retains the entire copyright notice properly, and       */
/*  displays or retains this license in the documentation and/or other        */
/*  materials provided with the distribution.                                 */
/*                                                                            */
/*  Redistribution and use in source form or binary form, with modification,  */
/*  the whole program or a portion of it, are permitted provided that one     */
/*  declares that it is modified software and not original, and displays or   */
/*  describes any modifications in the documentation and/or other materials   */
/*  provided with the distribution.                                           */
/*                                                                            */
/*  This software and the name of the author must not be used to endorse or   */
/*  promote products derived from this software without prior written         */
/*  permission.                                                               */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/*  EMail   abetakao@bea.hi-ho.ne.jp                                          */
/*          HCA00763@nifty.ne.jp                                              */
/*                                                                            */
/*  URL     http://www.bea.hi-ho.ne.jp/abetakao/                              */
/*                                                                            */
/******************************************************************************/


#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<time.h>


#include	<X11/Intrinsic.h>
#include	<X11/StringDefs.h>
#include	<X11/Shell.h>

#include	<X11/Xaw/AsciiText.h>
#include	<X11/Xaw/Box.h>
#include	<X11/Xaw/Command.h>
#include	<X11/Xaw/Form.h>
#include	<X11/Xaw/Label.h>


#include	"xmemo.h"
#include	"xmemodata.h"
#include	"xmemofunc.h"


/******************************************************************************/
/*  Local variables                                                           */
/******************************************************************************/


static	Widget	wDlgOptionProfile ;
static	Widget	wDlgOptionProfileForm ;
static	Widget	wDlgOptionProfileEdit ;
static	Widget	wDlgOptionProfileLbl [ 2 ] ;
static	Widget	wDlgOptionProfileTxtNickname ;
static	Widget	wDlgOptionProfileTxtGroupname ;
static	Widget	wDlgOptionProfileCmdClose ;
static	Widget	wDlgOptionProfileCmdCancel ;


/******************************************************************************/
/*  OptionProfileClose_Click                                                  */
/******************************************************************************/

static void OptionProfileClose_Click ( Widget w, caddr_t pClientData, caddr_t pCallData )
{

int 	n ;
Arg 	xArg [ 10 ] ;
char	*pTxtNickname, *pTxtGroupname ;

	if ( bMemoTraceLog ) xmemo_TraceLog ( "OptionProfileClose_Click", NULL ) ;

	xmemo_PopdownDialog_OptionProfile ( ) ;

	/* Input */

	n = 0 ;
	XtSetArg ( xArg[n], XtNstring, &pTxtNickname ) ; n ++ ;
	XtGetValues ( wDlgOptionProfileTxtNickname, xArg, n ) ;

	if ( pTxtNickname != NULL ) {
		strncpy ( sMemoNickname, pTxtNickname, MAXLEN_NICKNAME ) ;
		sMemoNickname[MAXLEN_NICKNAME-1] = 0 ;
	} else {
		sMemoNickname[0] = 0 ;
	}

	n = 0 ;
	XtSetArg ( xArg[n], XtNstring, &pTxtGroupname ) ; n ++ ;
	XtGetValues ( wDlgOptionProfileTxtGroupname, xArg, n ) ;

	if ( pTxtGroupname != NULL ) {
		strncpy ( sMemoGroupname, pTxtGroupname, MAXLEN_GROUPNAME ) ;
		sMemoGroupname[MAXLEN_GROUPNAME-1] = 0 ;
	} else {
		sMemoGroupname[0] = 0 ;
	}

	/* Save */

	xmemo_SaveOptionProfile ( ) ;

	/* Notify */

	xmemo_NetNotify_Broadcast ( XMEMO_NET_NOTIFY_CHANGE ) ;

}

static XtCallbackRec pOptionProfileClose_CallbackRec [ ] =
{
	{ (XtCallbackProc) OptionProfileClose_Click, NULL },
	{ NULL, NULL }
} ;

/******************************************************************************/
/*  OptionProfileCancel_Click                                                 */
/******************************************************************************/

static void OptionProfileCancel_Click ( Widget w, caddr_t pClientData, caddr_t pCallData )
{

zChainType	*pChain ;
zNetworkType	*pNetwork ;

	if ( bMemoTraceLog ) xmemo_TraceLog ( "OptionProfileCancel_Click", NULL ) ;

	xmemo_PopdownDialog_OptionProfile ( ) ;

}

static XtCallbackRec pOptionProfileCancel_CallbackRec [ ] =
{
	{ (XtCallbackProc) OptionProfileCancel_Click, NULL },
	{ NULL, NULL }
} ;

/******************************************************************************/
/*  xmemo_InitDialog_OptionProfile                                            */
/******************************************************************************/

int xmemo_InitDialog_OptionProfile ( )
{

int 	n ;
Arg 	xArg [ 16 ] ;

	if ( bMemoTraceLog ) xmemo_TraceLog ( "xmemo_InitDialog_OptionProfile", NULL ) ;

	/* Dialog shell */

	n = 0 ;
	XtSetArg ( xArg[n], XtNtitle , "xmemo - Option [ Profile ]" ) ; n ++ ;
	XtSetArg ( xArg[n], XtNwidth , 400 ) ; n ++ ;
	XtSetArg ( xArg[n], XtNheight, 100 ) ; n ++ ;
	wDlgOptionProfile = XtCreatePopupShell ( "DlgOptProfile", transientShellWidgetClass, wTop, xArg, n ) ;
	                                  
	/* Main form */

	wDlgOptionProfileForm = XtCreateManagedWidget ( "Form", formWidgetClass, wDlgOptionProfile, xArg, 0 ) ;

	/* Edit area */

	n = 0 ;
	/*XtSetArg ( xArg[n], XtNwidth , 450 ) ; n ++ ;*/
	XtSetArg ( xArg[n], XtNtop   , XtChainTop    ) ; n ++ ;
	XtSetArg ( xArg[n], XtNbottom, XtChainTop    ) ; n ++ ;
	XtSetArg ( xArg[n], XtNleft  , XtChainLeft   ) ; n ++ ;
	XtSetArg ( xArg[n], XtNright , XtChainRight  ) ; n ++ ;
	wDlgOptionProfileEdit = XtCreateManagedWidget ( "Body", formWidgetClass, wDlgOptionProfileForm, xArg, n ) ;

	n = 0 ;
	XtSetArg ( xArg[n], XtNwidth   , 100 ) ; n ++ ;
	XtSetArg ( xArg[n], XtNlabel   , "Nick name" ) ; n ++ ;
	XtSetArg ( xArg[n], XtNjustify , XtJustifyLeft ) ; n ++ ;
	XtSetArg ( xArg[n], XtNborderWidth, 0 ) ; n ++ ;	
	XtSetArg ( xArg[n], XtNtop   , XtChainTop    ) ; n ++ ;
	XtSetArg ( xArg[n], XtNleft  , XtChainLeft   ) ; n ++ ;
	XtSetArg ( xArg[n], XtNright , XtChainLeft   ) ; n ++ ;
	wDlgOptionProfileLbl[0] = XtCreateManagedWidget ( "Label", labelWidgetClass, wDlgOptionProfileEdit, xArg, n ) ;

	n = 0 ;
	XtSetArg ( xArg[n], XtNwidth    , 200 ) ; n ++ ;
	XtSetArg ( xArg[n], XtNtop      , XtChainTop    ) ; n ++ ;
	XtSetArg ( xArg[n], XtNleft     , XtChainLeft   ) ; n ++ ;
	XtSetArg ( xArg[n], XtNright    , XtChainLeft   ) ; n ++ ;
	XtSetArg ( xArg[n], XtNfromVert , NULL ) ; n ++ ;
	XtSetArg ( xArg[n], XtNfromHoriz, wDlgOptionProfileLbl[0] ) ; n ++ ;	
	XtSetArg ( xArg[n], XtNeditType , XawtextEdit   ) ; n ++ ;	
	wDlgOptionProfileTxtNickname = XtCreateManagedWidget ( "Text", asciiTextWidgetClass, wDlgOptionProfileEdit, xArg, n ) ;

	n = 0 ;
	XtSetArg ( xArg[n], XtNwidth   , 100 ) ; n ++ ;
	XtSetArg ( xArg[n], XtNlabel   , "Group" ) ; n ++ ;
	XtSetArg ( xArg[n], XtNjustify , XtJustifyLeft ) ; n ++ ;
	XtSetArg ( xArg[n], XtNborderWidth, 0 ) ; n ++ ;	
	XtSetArg ( xArg[n], XtNtop   , XtChainTop    ) ; n ++ ;
	XtSetArg ( xArg[n], XtNleft  , XtChainLeft   ) ; n ++ ;
	XtSetArg ( xArg[n], XtNright , XtChainLeft   ) ; n ++ ;
	XtSetArg ( xArg[n], XtNfromVert, wDlgOptionProfileLbl[0] ) ; n ++ ;
	wDlgOptionProfileLbl[1] = XtCreateManagedWidget ( "Label", labelWidgetClass, wDlgOptionProfileEdit, xArg, n ) ;

	n = 0 ;
	XtSetArg ( xArg[n], XtNwidth    , 200 ) ; n ++ ;
	XtSetArg ( xArg[n], XtNtop      , XtChainTop    ) ; n ++ ;
	XtSetArg ( xArg[n], XtNleft     , XtChainLeft   ) ; n ++ ;
	XtSetArg ( xArg[n], XtNright    , XtChainLeft   ) ; n ++ ;
	XtSetArg ( xArg[n], XtNfromVert , wDlgOptionProfileTxtNickname ) ; n ++ ;
	XtSetArg ( xArg[n], XtNfromHoriz, wDlgOptionProfileLbl[1] ) ; n ++ ;	
	XtSetArg ( xArg[n], XtNeditType , XawtextEdit   ) ; n ++ ;	
	wDlgOptionProfileTxtGroupname = XtCreateManagedWidget ( "Text", asciiTextWidgetClass, wDlgOptionProfileEdit, xArg, n ) ;

	/* Command button ( Bottom ) */

	n = 0 ;
	XtSetArg ( xArg[n], XtNlabel    , "OK" ) ; n ++ ;
	XtSetArg ( xArg[n], XtNtop      , XtChainBottom ) ; n ++ ;
	XtSetArg ( xArg[n], XtNbottom   , XtChainBottom ) ; n ++ ;
	XtSetArg ( xArg[n], XtNleft     , XtChainLeft   ) ; n ++ ;
	XtSetArg ( xArg[n], XtNright    , XtChainLeft   ) ; n ++ ;
	XtSetArg ( xArg[n], XtNfromVert , wDlgOptionProfileEdit ) ; n ++ ;
	XtSetArg ( xArg[n], XtNcallback , pOptionProfileClose_CallbackRec ) ; n ++ ;
	wDlgOptionProfileCmdClose = XtCreateManagedWidget ( "CmdButton", commandWidgetClass, wDlgOptionProfileForm, xArg, n ) ;

	n = 0 ;
	XtSetArg ( xArg[n], XtNlabel    , "Cancel" ) ; n ++ ;
	XtSetArg ( xArg[n], XtNtop      , XtChainBottom ) ; n ++ ;
	XtSetArg ( xArg[n], XtNbottom   , XtChainBottom ) ; n ++ ;
	XtSetArg ( xArg[n], XtNleft     , XtChainLeft   ) ; n ++ ;
	XtSetArg ( xArg[n], XtNright    , XtChainLeft   ) ; n ++ ;
	XtSetArg ( xArg[n], XtNfromVert , wDlgOptionProfileEdit ) ; n ++ ;
	XtSetArg ( xArg[n], XtNfromHoriz, wDlgOptionProfileCmdClose ) ; n ++ ;
	XtSetArg ( xArg[n], XtNcallback , pOptionProfileCancel_CallbackRec ) ; n ++ ;
	wDlgOptionProfileCmdCancel = XtCreateManagedWidget ( "CmdButton", commandWidgetClass, wDlgOptionProfileForm, xArg, n ) ;

	return RC_SUCCESS ;

}

/******************************************************************************/
/*  xmemo_PopupDialog_OptionProfile                                           */
/******************************************************************************/

int xmemo_PopupDialog_OptionProfile ( )
{

int 	n ;
Arg 	xArg [ 5 ] ;

	if ( bMemoTraceLog ) xmemo_TraceLog ( "xmemo_PopupDialog_OptionProfile", NULL ) ;

	n = 0 ;
	XtSetArg ( xArg[n], XtNstring, sMemoNickname ) ; n ++ ;
	XtSetValues ( wDlgOptionProfileTxtNickname, xArg, n ) ;

	n = 0 ;
	XtSetArg ( xArg[n], XtNstring, sMemoGroupname ) ; n ++ ;
	XtSetValues ( wDlgOptionProfileTxtGroupname, xArg, n ) ;

	XtPopup ( wDlgOptionProfile, XtGrabExclusive ) ;

	return RC_SUCCESS ;
	
}

/******************************************************************************/
/*  xmemo_PopdownDialog_OptionProfile                                         */
/******************************************************************************/

int xmemo_PopdownDialog_OptionProfile ( )
{

	if ( bMemoTraceLog ) xmemo_TraceLog ( "xmemo_PopdownDialog_OptionProfile", NULL ) ;

	XtPopdown ( wDlgOptionProfile ) ;

	return RC_SUCCESS ;

}
