/******************************************************************************/
/*                                                                            */
/*  <<< Copyright >>>                                                         */
/*                                                                            */
/*  The "xmemo" software is copyright 1999 by Takao Abe.                      */
/*  All rights reserved.                                                      */
/*                                                                            */
/*  The "xmemo" software refers to the source code, the binary code,          */
/*  the documents, and all derivatives based on this software under           */
/*  the copyright law, that is a software containing a program or a           */
/*  portion of a program, regardless of modifications and/or translations.    */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/*  <<< Licence and redistribution >>>                                        */
/*                                                                            */
/*  The source code and the binary code are free software.                    */
/*  The use and redistribution of source code and binary code are permitted   */
/*  provided that the following conditions are met:                           */
/*  You may use the software for private use and commercial use,              */
/*  and may modify your copy or copies of the software or any portion of it.  */
/*                                                                            */
/*  THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESSED OR IMPLIED            */
/*  WARRANTIES OF ANY KIND, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED        */
/*  WARRANTIES OF MERCHANTABLILITY AND FITNESS FOR A PARTICULAR PURPOSE.      */
/*  IN NO EVENT SHALL THE AUTHOR TAKAO ABE BE LIABLE FOR ANY DIRECT,          */
/*  INDIRECT, GENERAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES           */
/*  ( INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR       */
/*  SERVICES; LOSS OF USE, DATA OR PROFITS; OR BUSINESS INTERRUPTION )        */
/*  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,       */
/*  STRICT LIABILITY, OR TORT ( INCLUDING NEGLIGENCE OR OTHERWISE )           */
/*  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED       */
/*  OF THE POSSIBILITY OF SUCH DAMAGE.                                        */
/*                                                                            */
/*  Redistribution and use in source form or binary form, with or without     */
/*  modification, the whole program or a portion of it, are permitted         */
/*  provided that one retains the entire copyright notice properly, and       */
/*  displays or retains this license in the documentation and/or other        */
/*  materials provided with the distribution.                                 */
/*                                                                            */
/*  Redistribution and use in source form or binary form, with modification,  */
/*  the whole program or a portion of it, are permitted provided that one     */
/*  declares that it is modified software and not original, and displays or   */
/*  describes any modifications in the documentation and/or other materials   */
/*  provided with the distribution.                                           */
/*                                                                            */
/*  This software and the name of the author must not be used to endorse or   */
/*  promote products derived from this software without prior written         */
/*  permission.                                                               */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/*  EMail   abetakao@bea.hi-ho.ne.jp                                          */
/*          HCA00763@nifty.ne.jp                                              */
/*                                                                            */
/*  URL     http://www.bea.hi-ho.ne.jp/abetakao/                              */
/*                                                                            */
/******************************************************************************/


#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<time.h>


#include	<X11/Intrinsic.h>


#include	"xmemo.h"
#include	"xmemodata.h"
#include	"xmemofunc.h"


/******************************************************************************/
/*  Local constant definition                                                 */
/******************************************************************************/

#define		ISO_CHARSET_NULL    	' '
#define		ISO_CHARSET_ASCII   	'A'
#define		ISO_CHARSET_KATAKANA	'J'
#define		ISO_CHARSET_KANJI   	'K'

/******************************************************************************/
/*  xmemo_CsvToken                                                            */
/******************************************************************************/

int xmemo_CsvToken ( char *pToken[ ], int n, char *pStr )
{

int 	i, j ;

	i = 0 ;
	while ( i < n  &&  *pStr != 0 ) {
		pToken[i++] = pStr ;
		while ( *pStr != ','  &&  *pStr != 0 ) pStr ++ ;
		if ( *pStr == ',' ) {
			*pStr = 0 ;
			pStr ++ ;
		}
	}

	for ( j = i ; j < n ; j ++ ) pToken[j] = NULL ;

	return i ;

}

/******************************************************************************/
/*  xmemo_CheckNumeric                                                        */
/******************************************************************************/

int xmemo_CheckNumeric ( char *pStr )
{


char	*pSpace ;
int 	iValue ;

	if ( pStr == NULL ) return FALSE ;

	/* Skip leading spaces */
	while ( *pStr == ' ' ) pStr ++ ;
	if ( *pStr == 0 ) return FALSE ;

	/* Check trailing spaces */
	pSpace = strchr ( pStr, ' ' ) ;
	if ( pSpace != NULL ) {
		while ( *pSpace == ' ' ) pSpace ++ ;
		if ( *pSpace != 0 ) return FALSE ;
	}

	/* Check numeric */
	if ( strspn ( pStr, " 0123456789" ) != strlen ( pStr ) ) return FALSE ;

	return TRUE ;

}

/******************************************************************************/
/*  xmemo_CheckNumericRange                                                   */
/******************************************************************************/

int xmemo_CheckNumericRange ( char *pStr, int iLow, int iHigh )
{

char	*pSpace ;
int 	iValue ;

	if ( pStr == NULL ) return FALSE ;

	/* Skip leading spaces */
	while ( *pStr == ' ' ) pStr ++ ;
	if ( *pStr == 0 ) return FALSE ;

	/* Check trailing spaces */
	pSpace = strchr ( pStr, ' ' ) ;
	if ( pSpace != NULL ) {
		while ( *pSpace == ' ' ) pSpace ++ ;
		if ( *pSpace != 0 ) return FALSE ;
	}

	/* Check numeric */
	if ( strspn ( pStr, " 0123456789" ) != strlen ( pStr ) ) return FALSE ;

	/* Check range */
	iValue = atoi ( pStr ) ;
	if ( iValue < iLow || iHigh < iValue )  return FALSE ;

	return TRUE ;

}

/******************************************************************************/
/*  kanji_jis_to_shiftjis                                                     */
/******************************************************************************/

void kanji_jis_to_shiftjis ( char *pShiftJIS, char *pJIS )
{

unsigned	char	hibyte, lobyte ;

	hibyte = pJIS[0] & 0x7F ;
	lobyte = pJIS[1] & 0x7F ;

	if ( ( hibyte & 1 ) == 0 ) {
		hibyte = ( hibyte >> 1 ) - 1 ;
		lobyte += 0x7D ;
	} else {
		hibyte = ( hibyte >> 1 ) ;
		lobyte += 0x1F ;
	}
	if ( lobyte >= 0x7F ) lobyte ++ ;
	if ( hibyte < 0x2F ) {
		hibyte += 0x71 ;
	} else {
		hibyte += 0xB1 ;
	}

	pShiftJIS[0] = hibyte ;
	pShiftJIS[1] = lobyte ;

}

/******************************************************************************/
/*  kanji_euc_to_shiftjis                                                     */
/******************************************************************************/

void kanji_euc_to_shiftjis ( char *pShiftJIS, char *pEUC )
{

	kanji_jis_to_shiftjis ( pShiftJIS, pEUC ) ;

}

/******************************************************************************/
/*  kanji_shiftjis_to_jis                                                     */
/******************************************************************************/

void kanji_shiftjis_to_jis ( char *pJIS, char *pShiftJIS )
{

unsigned	char	hibyte, lobyte ;

	hibyte = pShiftJIS[0] ;
	lobyte = pShiftJIS[1] ;

	if ( hibyte < 0xA0 ) {
		hibyte -= 0x70 ;
	} else {
		hibyte -= 0xB0 ;
	}
	hibyte <<= 1 ;
	if ( lobyte >= 0x80 ) lobyte -- ;
	if ( lobyte >= 0x9E ) {
		lobyte -= 0x7D ;
	} else {
		lobyte -= 0x1F ;
		hibyte -- ;
	}

	pJIS[0] = hibyte ;
	pJIS[1] = lobyte ;

}

/******************************************************************************/
/*  kanji_shiftjis_to_euc                                                     */
/******************************************************************************/

void kanji_shiftjis_to_euc ( char *pEUC, char *pShiftJIS )
{

	kanji_shiftjis_to_jis ( pEUC, pShiftJIS ) ;

	pEUC[0] |= 0x80 ;
	pEUC[1] |= 0x80 ;

}

/******************************************************************************/
/*  ShiftJIS_CRLF                                                             */
/******************************************************************************/

int ShiftJIS_CRLF ( char *pDest, char *pSource, int iLen, char *pCR, char *pLF )
{

int 	i = 0 ;
unsigned	char	ch ;

	while ( i < iLen - 1  &&  *pSource != 0 ) {

		ch = *pSource ;
		if ( ch == 0x0D ) {
			if ( pCR[0] != 0  &&  i < iLen - 1 ) {
				*pDest++ = pCR[0] ;
				i ++ ;
			}
			if ( pCR[1] != 0  &&  i < iLen - 1 ) {
				*pDest++ = pCR[1] ;
				i ++ ;
			}
			pSource ++ ;
		} else if ( ch == 0x0A ) {
			if ( pLF[0] != 0  &&  i < iLen - 1 ) {
				*pDest++ = pLF[0] ;
				i ++ ;
			}
			if ( pLF[1] != 0  &&  i < iLen - 1 ) {
				*pDest++ = pLF[1] ;
				i ++ ;
			}
			pSource ++ ;
		} else {
			*pDest++ = *pSource++ ;
			i ++ ;
		}

	}
	*pDest = 0 ;

	return i ;

}

/******************************************************************************/
/*  ShiftJIS_To_JIS7                                                          */
/******************************************************************************/

int ShiftJIS_To_JIS7 ( char *pDest, char *pSource, int iLen, char cCR, char cLF )
{

int 	i = 0 ;
int 	bKanji = FALSE ;
int 	bKatakana = FALSE ;
unsigned	char	ch ;

	while ( i < iLen - 1  &&  *pSource != 0 ) {
		ch = *pSource ;
		if ( ch == 0x0D ) {
			if ( cCR != 0 ) {
				*pDest++ = cCR ;
				i ++ ;
			}
			pSource ++ ;
		} else if ( ch == 0x0A ) {
			if ( cLF != 0 ) {
				*pDest++ = cLF ;
				i ++ ;
			}
			pSource ++ ;
		} else if ( ch <= 0x7F ) {
			if ( bKanji  ||  bKatakana ) {
				if ( i + 4 >= iLen ) break ;
				bKanji = FALSE ;
				bKatakana = FALSE ;
				*pDest = 0x1B ;
				*pDest = 0x28 ;
				*pDest = 0x4A ;
				i += 3 ;
			}
			*pDest++ = *pSource++ ;
			i ++ ;
		} else if ( 0x80 <= ch && ch <= 0x9F || 0xE0 <= ch && ch <= 0xFF ) {
			/* Kanji */
			if ( ! bKanji ) {
				if ( i + 5 >= iLen ) break ;
				bKanji = TRUE ;
				bKatakana = FALSE ;
				*pDest = 0x1B ;
				*pDest = 0x24 ;
				*pDest = 0x42 ;
				i += 3 ;
			}
			if ( i + 2 >= iLen  || pSource[1] == 0 ) break ;
			kanji_shiftjis_to_jis ( pDest, pSource ) ;
			pDest += 2 ;
			pSource += 2 ;
			i += 2 ;
		} else {
			/* Katakana ( Half size ) */
			if ( ! bKatakana ) {
				if ( i + 4 >= iLen ) break ;
				bKanji = FALSE ;
				bKatakana = TRUE ;
				*pDest = 0x1B ;
				*pDest = 0x28 ;
				*pDest = 0x49 ;
				i += 3 ;
			}
			*pDest++ = *pSource++ ;
			i ++ ;
		}
	}
	*pDest = 0 ;

	return i ;

}

/******************************************************************************/
/*  ShiftJIS_To_JIS8                                                          */
/******************************************************************************/

int ShiftJIS_To_JIS8 ( char *pDest, char *pSource, int iLen, char cCR, char cLF )
{

int 	i = 0 ;
int 	bKanji = FALSE ;
unsigned	char	ch ;

	while ( i < iLen - 1  &&  *pSource != 0 ) {
		ch = *pSource ;
		if ( ch == 0x0D ) {
			if ( cCR != 0 ) {
				*pDest++ = cCR ;
				i ++ ;
			}
			pSource ++ ;
		} else if ( ch == 0x0A ) {
			if ( cLF != 0 ) {
				*pDest++ = cLF ;
				i ++ ;
			}
			pSource ++ ;
		} else if ( ch <= 0x7F ) {
			if ( bKanji ) {
				if ( i + 4 >= iLen ) break ;
				bKanji = FALSE ;
				*pDest = 0x1B ;
				*pDest = 0x28 ;
				*pDest = 0x4A ;
				i += 3 ;
			}
			*pDest++ = *pSource++ ;
			i ++ ;
		} else if ( 0x80 <= ch && ch <= 0x9F || 0xE0 <= ch && ch <= 0xFF ) {
			/* Kanji */
			if ( ! bKanji ) {
				if ( i + 5 >= iLen ) break ;
				bKanji = TRUE ;
				*pDest = 0x1B ;
				*pDest = 0x24 ;
				*pDest = 0x42 ;
				i += 3 ;
			}
			if ( i + 2 >= iLen  || pSource[1] == 0 ) break ;
			kanji_shiftjis_to_jis ( pDest, pSource ) ;
			pDest += 2 ;
			pSource += 2 ;
			i += 2 ;
		} else {
			/* Katakana ( Half size ) */
			if ( bKanji ) {
				if ( i + 4 >= iLen ) break ;
				bKanji = FALSE ;
				*pDest = 0x1B ;
				*pDest = 0x28 ;
				*pDest = 0x4A ;
				i += 3 ;
			}
			*pDest++ = *pSource++ ;
			i ++ ;
		}
	}
	*pDest = 0 ;

	return i ;

}

/******************************************************************************/
/*  ShiftJIS_To_EUC                                                           */
/******************************************************************************/

int ShiftJIS_To_EUC ( char *pDest, char *pSource, int iLen, char cCR, char cLF )
{

int 	i = 0 ;
unsigned	char	ch ;

	while ( i < iLen - 1  &&  *pSource != 0 ) {
		ch = *pSource ;
		if ( ch == 0x0D ) {
			if ( cCR != 0 ) {
				*pDest++ = cCR ;
				i ++ ;
			}
			pSource ++ ;
		} else if ( ch == 0x0A ) {
			if ( cLF != 0 ) {
				*pDest++ = cLF ;
				i ++ ;
			}
			pSource ++ ;
		} else if ( ch <= 0x7F ) {
			*pDest++ = *pSource++ ;
			i ++ ;
		} else if ( 0x80 <= ch && ch <= 0x9F || 0xE0 <= ch && ch <= 0xFF ) {
			/* Kanji */
			if ( i + 2 >= iLen  || pSource[1] == 0 ) break ;
			kanji_shiftjis_to_euc ( pDest, pSource ) ;
			pDest += 2 ;
			pSource += 2 ;
			i += 2 ;
		} else {
			/* Katakana ( Half size ) */
			if ( i + 2 >= iLen ) break ;
			*pDest++ = 0x8E ;
			*pDest++ = *pSource++ ;
			i += 2 ;
		}
	}
	*pDest = 0 ;

	return i ;

}

/******************************************************************************/
/*  ISO_DefaultDesignation_JIS                                                */
/******************************************************************************/

char* ISO_DefaultDesignation_JIS ( )
{

static	char	sDefaultDesignation [ ] = {
	ISO_CHARSET_ASCII, 
	ISO_CHARSET_KATAKANA, 
	ISO_CHARSET_NULL, 
	ISO_CHARSET_NULL,
	0 } ;

	return sDefaultDesignation ;

}

/******************************************************************************/
/*  ISO_DefaultDesignation_EUC                                                */
/******************************************************************************/

char* ISO_DefaultDesignation_EUC ( )
{

static	char	sDefaultDesignation [ ] = {
	ISO_CHARSET_ASCII, 
	ISO_CHARSET_KANJI, 
	ISO_CHARSET_KATAKANA, 
	ISO_CHARSET_NULL,
	0 } ;

	return sDefaultDesignation ;

}

/******************************************************************************/
/*  ISO_To_ShiftJIS                                                           */
/******************************************************************************/

int ISO_To_ShiftJIS ( char *pDest, char *pSource, int iLen, char *pDesignation, char *pNewLine )
{

unsigned	char	ch ;
int 	i, iSkip, bEndOfData ;
char	*pGL, *pGR, *pSS ;
char	cG0, cG1, cG2, cG3 ;

	/* Default invocation */

	pGL = &cG0 ;
	pGR = &cG1 ;
	pSS = NULL ;

	/* Default designation */

	cG0 = pDesignation[0] ;
	cG1 = pDesignation[1] ;
	cG2 = pDesignation[2] ;
	cG3 = pDesignation[3] ;

	/* JIS or EUC to Shift-JIS with CRLF conversion */

	i = 0 ;
	bEndOfData = FALSE ;
	while ( i < iLen - 1  &&  *pSource != 0  &&  ! bEndOfData ) {

		ch = *pSource ;

		if ( ch == 0x1B ) {
			/* ESC */
			iSkip = 1 ;
			ch = pSource[1] ;
			if ( 0x20 <= ch  &&  ch <= 0x7F ) iSkip ++ ;
			switch ( ch ) {
			case 0x24 :
				ch = pSource[2] ;
				if ( 0x20 <= ch  &&  ch <= 0x7F ) iSkip ++ ;
				switch ( ch ) {
				case 0x40 : cG0 = ISO_CHARSET_KANJI ; break ;
				case 0x42 : cG0 = ISO_CHARSET_KANJI ; break ;
				case 0x28 :
					ch = pSource[3] ;
					if ( 0x20 <= ch  &&  ch <= 0x7F ) iSkip ++ ;
					switch ( ch ) {
					case 0x40 : cG0 = ISO_CHARSET_KANJI ; break ;
					case 0x42 : cG0 = ISO_CHARSET_KANJI ; break ;
					}
					break ;
				case 0x29 :
					ch = pSource[3] ;
					if ( 0x20 <= ch  &&  ch <= 0x7F ) iSkip ++ ;
					switch ( ch ) {
					case 0x40 : cG1 = ISO_CHARSET_KANJI ; break ;
					case 0x42 : cG1 = ISO_CHARSET_KANJI ; break ;
					}
					break ;
				case 0x2A :
					ch = pSource[3] ;
					if ( 0x20 <= ch  &&  ch <= 0x7F ) iSkip ++ ;
					switch ( ch ) {
					case 0x40 : cG2 = ISO_CHARSET_KANJI ; break ;
					case 0x42 : cG2 = ISO_CHARSET_KANJI ; break ;
					}
					break ;
				case 0x2B :
					ch = pSource[3] ;
					if ( 0x20 <= ch  &&  ch <= 0x7F ) iSkip ++ ;
					switch ( ch ) {
					case 0x40 : cG3 = ISO_CHARSET_KANJI ; break ;
					case 0x42 : cG3 = ISO_CHARSET_KANJI ; break ;
					}
					break ;
				}
				break ;
			case 0x28 :
				ch = pSource[2] ;
				if ( 0x20 <= ch  &&  ch <= 0x7F ) iSkip ++ ;
				switch ( ch ) {
				case 0x42 : cG0 = ISO_CHARSET_ASCII    ; break ;
				case 0x49 : cG0 = ISO_CHARSET_KATAKANA ; break ;
				case 0x4A : cG0 = ISO_CHARSET_ASCII    ; break ;
				}
				break ;
			case 0x29 :
				ch = pSource[2] ;
				if ( 0x20 <= ch  &&  ch <= 0x7F ) iSkip ++ ;
				switch ( ch ) {
				case 0x42 : cG1 = ISO_CHARSET_ASCII    ; break ;
				case 0x49 : cG1 = ISO_CHARSET_KATAKANA ; break ;
				case 0x4A : cG1 = ISO_CHARSET_ASCII    ; break ;
				}
				break ;
			case 0x2A :
				ch = pSource[2] ;
				if ( 0x20 <= ch  &&  ch <= 0x7F ) iSkip ++ ;
				switch ( ch ) {
				case 0x42 : cG2 = ISO_CHARSET_ASCII    ; break ;
				case 0x49 : cG2 = ISO_CHARSET_KATAKANA ; break ;
				case 0x4A : cG2 = ISO_CHARSET_ASCII    ; break ;
				}
				break ;
			case 0x2B :
				ch = pSource[2] ;
				if ( 0x20 <= ch  &&  ch <= 0x7F ) iSkip ++ ;
				switch ( ch ) {
				case 0x42 : cG3 = ISO_CHARSET_ASCII    ; break ;
				case 0x49 : cG3 = ISO_CHARSET_KATAKANA ; break ;
				case 0x4A : cG3 = ISO_CHARSET_ASCII    ; break ;
				}
				break ;
			case 0x6E : pGL = &cG2 ; break ;
			case 0x6F : pGL = &cG3 ; break ;
			case 0x7C : pGR = &cG3 ; break ;
			case 0x7D : pGR = &cG2 ; break ;
			case 0x7E : pGR = &cG1 ; break ;
			}
			pSource += iSkip ;
			pSS = NULL ;

		} else if ( ch <= 0x0A ) {
			/* LF */
			if ( pNewLine[0] != 0  &&  i < iLen - 1 ) {
				*pDest++ = pNewLine[0] ;
				i ++ ;
			}
			if ( pNewLine[1] != 0  &&  i < iLen - 1 ) {
				*pDest++ = pNewLine[1] ;
				i ++ ;
			}
			pSource ++ ;
			pSS = NULL ;

		} else if ( ch <= 0x0D ) {
			/* CR */
			*pDest++ = *pSource++ ;
			i ++ ;
			pSS = NULL ;

		} else if ( ch <= 0x0E ) {
			/* SO ( LS1 ) */
			pGL = &cG1 ;
			pSource ++ ;
			pSS = NULL ;

		} else if ( ch <= 0x0F ) {
			/* SI ( LS0 ) */
			pGL = &cG0 ;
			pSource ++ ;
			pSS = NULL ;

		} else if ( ch <= 0x1F ) {
			/* C0 ( Other ) */
			*pDest++ = *pSource++ ;
			i ++ ;
			pSS = NULL ;

		} else if ( ch <= 0x7F ) {
			/* GL */
			switch ( *pGL ) {
			case ISO_CHARSET_KANJI :
				if ( i + 1 < iLen - 1  &&  pSource[1] == 0 ) {
					bEndOfData = TRUE ;
					break ;
				}
				kanji_jis_to_shiftjis ( pDest, pSource ) ;
				pDest += 2 ;
				pSource += 2 ;
				i += 2 ;
				break ;
			case ISO_CHARSET_KATAKANA :
				*pDest++ = *pSource++ | 0x80 ;
				i ++ ;
				break ;
			default :
				*pDest++ = *pSource++ ;
				i ++ ;
				break ;
			}
			pSS = NULL ;

		} else if ( ch <= 0x8E ) {
			/* C0 ( SS2 ) */
			pSS = &cG2 ;
			pSource ++ ;

		} else if ( ch <= 0x8F ) {
			/* C0 ( SS3 ) */
			pSS = &cG3 ;
			pSource ++ ;

		} else if ( ch <= 0x9F ) {
			/* C1 */
			*pDest++ = *pSource++ ;
			i ++ ;
			pSS = NULL ;

		} else if ( pSS != NULL ) {
			/* GR ( Single shift ) */
			switch ( *pSS ) {
			case ISO_CHARSET_KANJI :
				if ( i + 1 < iLen - 1  &&  pSource[1] == 0 ) {
					bEndOfData = TRUE ;
					break ;
				}
				kanji_jis_to_shiftjis ( pDest, pSource ) ;
				pDest += 2 ;
				pSource += 2 ;
				i += 2 ;
				break ;
			case ISO_CHARSET_KATAKANA :
				*pDest++ = *pSource++ ;
				i ++ ;
				break ;
			default :
				*pDest++ = *pSource++ & 0x7F ;
				i ++ ;
				break ;
			}
			pSS = NULL ;

		} else {
			/* GR */
			switch ( *pGR ) {
			case ISO_CHARSET_KANJI :
				if ( i + 1 < iLen - 1  &&  pSource[1] == 0 ) {
					bEndOfData = TRUE ;
					break ;
				}
				kanji_jis_to_shiftjis ( pDest, pSource ) ;
				pDest += 2 ;
				pSource += 2 ;
				i += 2 ;
				break ;
			case ISO_CHARSET_KATAKANA :
				*pDest++ = *pSource++ ;
				i ++ ;
				break ;
			default :
				*pDest++ = *pSource++ & 0x7F ;
				i ++ ;
				break ;
			}

		}

	}
	*pDest = 0 ;

	return i ;

}
