/******************************************************************************/
/*                                                                            */
/*  <<< Copyright >>>                                                         */
/*                                                                            */
/*  The "xmemo" software is copyright 1999 by Takao Abe.                      */
/*  All rights reserved.                                                      */
/*                                                                            */
/*  The "xmemo" software refers to the source code, the binary code,          */
/*  the documents, and all derivatives based on this software under           */
/*  the copyright law, that is a software containing a program or a           */
/*  portion of a program, regardless of modifications and/or translations.    */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/*  <<< Licence and redistribution >>>                                        */
/*                                                                            */
/*  The source code and the binary code are free software.                    */
/*  The use and redistribution of source code and binary code are permitted   */
/*  provided that the following conditions are met:                           */
/*  You may use the software for private use and commercial use,              */
/*  and may modify your copy or copies of the software or any portion of it.  */
/*                                                                            */
/*  THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESSED OR IMPLIED            */
/*  WARRANTIES OF ANY KIND, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED        */
/*  WARRANTIES OF MERCHANTABLILITY AND FITNESS FOR A PARTICULAR PURPOSE.      */
/*  IN NO EVENT SHALL THE AUTHOR TAKAO ABE BE LIABLE FOR ANY DIRECT,          */
/*  INDIRECT, GENERAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES           */
/*  ( INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR       */
/*  SERVICES; LOSS OF USE, DATA OR PROFITS; OR BUSINESS INTERRUPTION )        */
/*  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,       */
/*  STRICT LIABILITY, OR TORT ( INCLUDING NEGLIGENCE OR OTHERWISE )           */
/*  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED       */
/*  OF THE POSSIBILITY OF SUCH DAMAGE.                                        */
/*                                                                            */
/*  Redistribution and use in source form or binary form, with or without     */
/*  modification, the whole program or a portion of it, are permitted         */
/*  provided that one retains the entire copyright notice properly, and       */
/*  displays or retains this license in the documentation and/or other        */
/*  materials provided with the distribution.                                 */
/*                                                                            */
/*  Redistribution and use in source form or binary form, with modification,  */
/*  the whole program or a portion of it, are permitted provided that one     */
/*  declares that it is modified software and not original, and displays or   */
/*  describes any modifications in the documentation and/or other materials   */
/*  provided with the distribution.                                           */
/*                                                                            */
/*  This software and the name of the author must not be used to endorse or   */
/*  promote products derived from this software without prior written         */
/*  permission.                                                               */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/*  EMail   abetakao@bea.hi-ho.ne.jp                                          */
/*          HCA00763@nifty.ne.jp                                              */
/*                                                                            */
/*  URL     http://www.bea.hi-ho.ne.jp/abetakao/                              */
/*                                                                            */
/******************************************************************************/


#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<time.h>


#include	<X11/Intrinsic.h>


#include	"xmemo.h"
#include	"xmemodata.h"
#include	"xmemofunc.h"


/******************************************************************************/
/*  xmemo_ErrorLog                                                            */
/******************************************************************************/

void xmemo_ErrorLog ( char *sFunc, char *sReason, char *sRemarks )
{

time_t	t ;
struct	tm	*pNow ;
char	sDateTime [ 32 ] ;
char	fn [ MAXLEN_FILENAME ] ;
FILE	*fp ;

	time ( &t ) ;
	pNow = localtime ( &t ) ;

	sprintf ( sDateTime, "%04d/%02d/%02d  %02d:%02d:%02d",
	          pNow->tm_year+1900, pNow->tm_mon+1, pNow->tm_mday,
	          pNow->tm_hour, pNow->tm_min, pNow->tm_sec ) ;

	/* Write onto stderr */

	fprintf ( stderr, "%s", sDateTime ) ;
	if ( sFunc    != NULL ) fprintf ( stderr, "  %-30s", sFunc    ) ;
	else                    fprintf ( stderr, "  %-30s", " "      ) ;
	if ( sReason  != NULL ) fprintf ( stderr, "  %-30s", sReason  ) ;
	else                    fprintf ( stderr, "  %-30s", " "      ) ;
	if ( sRemarks != NULL ) fprintf ( stderr, "  %s\n" , sRemarks ) ;
	else                    fprintf ( stderr, "  %s\n" , " "      ) ;

	/* Write onto log file */

	strcpy ( fn, sMemoBaseDir ) ;
	strcat ( fn, "/" ) ;
	strcat ( fn, sMemoLogDir ) ;
	strcat ( fn, "/" ) ;
	strcat ( fn, sFilenameErrorLog ) ;
	fp = fopen ( fn, "a" ) ;
	if ( fp == NULL ) return ;

	fprintf ( fp, "%s", sDateTime ) ;
	if ( sFunc    != NULL ) fprintf ( fp, "  %-30s", sFunc    ) ;
	else                    fprintf ( fp, "  %-30s", " "      ) ;
	if ( sReason  != NULL ) fprintf ( fp, "  %-30s", sReason  ) ;
	else                    fprintf ( fp, "  %-30s", " "      ) ;
	if ( sRemarks != NULL ) fprintf ( fp, "  %s\n" , sRemarks ) ;
	else                    fprintf ( fp, "  %s\n" , " "      ) ;

	fclose ( fp ) ;

	return ;

}

/******************************************************************************/
/*  xmemo_TraceLog                                                            */
/******************************************************************************/

void xmemo_TraceLog ( char *sFunc, char *sLog )
{

time_t	t ;
struct	tm	*pNow ;
char	sDateTime [ 32 ] ;
char	fn [ MAXLEN_FILENAME ] ;
FILE	*fp ;

	time ( &t ) ;
	pNow = localtime ( &t ) ;

	sprintf ( sDateTime, "%04d/%02d/%02d  %02d:%02d:%02d",
	          pNow->tm_year+1900, pNow->tm_mon+1, pNow->tm_mday,
	          pNow->tm_hour, pNow->tm_min, pNow->tm_sec ) ;

	/* Write onto stderr */

	fprintf ( stdout, "%s", sDateTime ) ;
	if ( sFunc != NULL ) fprintf ( stdout, "  %-30s", sFunc ) ;
	else                 fprintf ( stdout, "  %-30s", " "   ) ;
	if ( sLog  != NULL ) fprintf ( stdout, "  %s\n" , sLog  ) ;
	else                 fprintf ( stdout, "  %s\n" , " "   ) ;

	/* Write onto log file */

	strcpy ( fn, sMemoBaseDir ) ;
	strcat ( fn, "/" ) ;
	strcat ( fn, sMemoLogDir ) ;
	strcat ( fn, "/" ) ;
	strcat ( fn, sFilenameTraceLog ) ;
	fp = fopen ( fn, "a" ) ;
	if ( fp == NULL ) return ;

	fprintf ( fp, "%s", sDateTime ) ;
	if ( sFunc != NULL ) fprintf ( fp, "  %-30s", sFunc ) ;
	else                 fprintf ( fp, "  %-30s", " "   ) ;
	if ( sLog  != NULL ) fprintf ( fp, "  %s\n" , sLog  ) ;
	else                 fprintf ( fp, "  %s\n" , " "   ) ;

	fclose ( fp ) ;

	return ;

}

/******************************************************************************/
/*  xmemo_TracePacket                                                         */
/******************************************************************************/

void xmemo_TracePacket ( char *pData, int n )
{

int 	i, j ;
char	sHex [ 4 ] ;
char	sDump [ 7 + 3 * 16 + 1 ] ;
char	sText [ 17 ] ;
char	fn [ MAXLEN_FILENAME ] ;
FILE	*fp ;

	strcpy ( fn, sMemoBaseDir ) ;
	strcat ( fn, "/" ) ;
	strcat ( fn, sMemoLogDir ) ;
	strcat ( fn, "/" ) ;
	strcat ( fn, sFilenameTraceLog ) ;
	fp = fopen ( fn, "a" ) ;

	for ( i = 0 ; i < n ; i += 16 ) {
		sprintf ( sDump, "%04X   ", i ) ;
		for ( j = i ; j - i < 16 && j < n ; j ++ ) {
			sprintf ( sHex, "%02X ", pData[j] & 0xFF ) ;
			strcat ( sDump, sHex ) ;
			if ( 0x20 <= pData[j]  &&  pData[j] <= 0x7E ) sText[j-i] = pData[j] ;
			else                                          sText[j-i] = ' ' ;
		}
		while ( j - i < 16 ) {
			strcat ( sDump, "   " ) ;
			sText[j-i] = ' ' ;
			j ++ ;
		}
		sText[16] = 0 ;
		fprintf ( stdout, "%s  [ %s ]\n", sDump, sText ) ;
		if ( fp != NULL ) fprintf ( fp, "%s  [ %s ]\n", sDump, sText ) ;
	}

	if ( fp != NULL ) fclose ( fp ) ;

}
