/******************************************************************************/
/*                                                                            */
/*  <<< Copyright >>>                                                         */
/*                                                                            */
/*  The "xmemo" software is copyright 1999 by Takao Abe.                      */
/*  All rights reserved.                                                      */
/*                                                                            */
/*  The "xmemo" software refers to the source code, the binary code,          */
/*  the documents, and all derivatives based on this software under           */
/*  the copyright law, that is a software containing a program or a           */
/*  portion of a program, regardless of modifications and/or translations.    */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/*  <<< Licence and redistribution >>>                                        */
/*                                                                            */
/*  The source code and the binary code are free software.                    */
/*  The use and redistribution of source code and binary code are permitted   */
/*  provided that the following conditions are met:                           */
/*  You may use the software for private use and commercial use,              */
/*  and may modify your copy or copies of the software or any portion of it.  */
/*                                                                            */
/*  THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESSED OR IMPLIED            */
/*  WARRANTIES OF ANY KIND, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED        */
/*  WARRANTIES OF MERCHANTABLILITY AND FITNESS FOR A PARTICULAR PURPOSE.      */
/*  IN NO EVENT SHALL THE AUTHOR TAKAO ABE BE LIABLE FOR ANY DIRECT,          */
/*  INDIRECT, GENERAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES           */
/*  ( INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR       */
/*  SERVICES; LOSS OF USE, DATA OR PROFITS; OR BUSINESS INTERRUPTION )        */
/*  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,       */
/*  STRICT LIABILITY, OR TORT ( INCLUDING NEGLIGENCE OR OTHERWISE )           */
/*  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED       */
/*  OF THE POSSIBILITY OF SUCH DAMAGE.                                        */
/*                                                                            */
/*  Redistribution and use in source form or binary form, with or without     */
/*  modification, the whole program or a portion of it, are permitted         */
/*  provided that one retains the entire copyright notice properly, and       */
/*  displays or retains this license in the documentation and/or other        */
/*  materials provided with the distribution.                                 */
/*                                                                            */
/*  Redistribution and use in source form or binary form, with modification,  */
/*  the whole program or a portion of it, are permitted provided that one     */
/*  declares that it is modified software and not original, and displays or   */
/*  describes any modifications in the documentation and/or other materials   */
/*  provided with the distribution.                                           */
/*                                                                            */
/*  This software and the name of the author must not be used to endorse or   */
/*  promote products derived from this software without prior written         */
/*  permission.                                                               */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/*  EMail   abetakao@bea.hi-ho.ne.jp                                          */
/*          HCA00763@nifty.ne.jp                                              */
/*                                                                            */
/*  URL     http://www.bea.hi-ho.ne.jp/abetakao/                              */
/*                                                                            */
/******************************************************************************/


#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<time.h>
#include	<errno.h>
#include	<sys/types.h>
#include	<sys/time.h>
#include	<sys/ioctl.h>
#include	<sys/socket.h>
#ifdef __sparc
#include	<sys/filio.h>
#endif
#include	<netdb.h>
#include	<netinet/in.h>
#include	<arpa/inet.h>


#include	<X11/Intrinsic.h>


#include	"xmemo.h"
#include	"xmemodata.h"
#include	"xmemofunc.h"


/******************************************************************************/
/*  Application protocol and format : IPMsg                                   */
/******************************************************************************/

/*  Header  */
#define IPMSG_VERSION       	1
#define IPMSG_DEFAULT_PORT  	2425

/*  Command  */
#define IPMSG_NOOPERATION   	0x00000000L

#define IPMSG_BR_ENTRY      	0x00000001L
#define IPMSG_BR_EXIT       	0x00000002L
#define IPMSG_ANSENTRY      	0x00000003L
#define IPMSG_BR_ABSENCE    	0x00000004L

#define IPMSG_BR_ISGETLIST  	0x00000010L
#define IPMSG_OKGETLIST     	0x00000011L
#define IPMSG_GETLIST       	0x00000012L
#define IPMSG_ANSLIST       	0x00000013L
#define IPMSG_BR_ISGETLIST2 	0x00000018L

#define IPMSG_SENDMSG       	0x00000020L
#define IPMSG_RECVMSG       	0x00000021L
#define IPMSG_READMSG       	0x00000030L
#define IPMSG_DELMSG        	0x00000031L

#define IPMSG_GETINFO       	0x00000040L
#define IPMSG_SENDINFO      	0x00000041L

#define IPMSG_GETABSENCEINFO	0x00000050L
#define IPMSG_SENDABSENCEINFO	0x00000051L

/*  Option for all command  */
#define IPMSG_ABSENCEOPT    	0x00000100L
#define IPMSG_SERVEROPT     	0x00000200L
#define IPMSG_DIALUPOPT     	0x00010000L

/*  Option for send command  */
#define IPMSG_SENDCHECKOPT  	0x00000100L
#define IPMSG_SECRETOPT     	0x00000200L
#define IPMSG_BROADCASTOPT  	0x00000400L
#define IPMSG_MULTICASTOPT  	0x00000800L
#define IPMSG_NOPOPUPOPT    	0x00001000L
#define IPMSG_AUTORETOPT    	0x00002000L
#define IPMSG_RETRYOPT      	0x00004000L
#define IPMSG_PASSWORDOPT   	0x00008000L
#define IPMSG_NOLOGOPT      	0x00020000L
#define IPMSG_NEWMUTIOPT    	0x00040000L
#define IPMSG_NOADDLISTOPT  	0x00080000L

/******************************************************************************/
/*  Local constant definition                                                 */
/******************************************************************************/

#define		XMEMO_IPMSG_TIMEOUT 	5
#define		XMEMO_IPMSG_RETRYOUT	3

/******************************************************************************/
/*  xmemo_DecodeMessage_Ipmsg                                                 */
/******************************************************************************/

static int xmemo_DecodeMessage_Ipmsg ( char *pDest, char *pSource, int iLen )
{

static	char	sCR [ ] = { 0x00, 0x00, 0x00 } ;
static	char	sLF [ ] = { 0x0A, 0x00, 0x00 } ;
int 	rc ;

	switch ( cMemoCharset ) {
	case CHARSET_EUC :
		rc = ShiftJIS_To_EUC ( pDest, pSource, iLen, 0, 0x0A ) ;
		break ;
	case CHARSET_JIS7 :
		rc = ShiftJIS_To_JIS7 ( pDest, pSource, iLen, 0, 0x0A ) ;
		break ;
	case CHARSET_JIS8 :
		rc = ShiftJIS_To_JIS8 ( pDest, pSource, iLen, 0, 0x0A ) ;
		break ;
	default :
		rc = ShiftJIS_CRLF ( pDest, pSource, iLen, sCR, sLF ) ;
		break ;
	}

	return rc ;

}

/******************************************************************************/
/*  xmemo_EncodeMessaage_Ipmsg                                                */
/******************************************************************************/

static int xmemo_EncodeMessaage_Ipmsg ( char *pDest, char *pSource, int iLen )
{

static	char	sCR [ ] = { 0x00, 0x00, 0x00 } ;
static	char	sLF [ ] = { 0x0A, 0x00, 0x00 } ;
int 	rc ;

	if ( pSource == NULL ) {
		*pDest = 0 ;
		return 0 ;
	}

	switch ( cMemoCharset ) {
	case CHARSET_EUC :
		rc = ISO_To_ShiftJIS ( pDest, pSource, iLen, ISO_DefaultDesignation_EUC( ), sLF ) ;
		break ;
	case CHARSET_JIS7 :
	case CHARSET_JIS8 :
		rc = ISO_To_ShiftJIS ( pDest, pSource, iLen, ISO_DefaultDesignation_JIS( ), sLF ) ;
		break ;
	default :
		rc = ShiftJIS_CRLF ( pDest, pSource, iLen, sCR, sLF ) ;
		break ;
	}


	return rc ;

}

/******************************************************************************/
/*  xmemo_IpmsgPacketNo                                                       */
/******************************************************************************/

static int xmemo_IpmsgPacketNo ( )
{

static	time_t	tPacket ;
static	int 	iPacket ;

time_t	tNow ;
struct	tm	*pNow ;

	time ( &tNow ) ;
	pNow = localtime ( &tNow ) ;
	if ( tNow != tPacket ) {
		tPacket = tNow ;
		iPacket = 0 ;
	} else {
		iPacket ++ ;
		if ( iPacket > 9999 ) iPacket = 0 ;
	}
	return ( pNow->tm_hour*3600 + pNow->tm_min*60 + pNow->tm_sec ) * 10000 + iPacket ;

}

/******************************************************************************/
/*  xmemo_IpmsgToken                                                          */
/******************************************************************************/

static int xmemo_IpmsgToken ( char *pToken[ ], int iTokenLen[ ], int n, char *pStr, int iLen )
{

int 	i = 0 ;
int 	nToken = 0 ;

	while ( nToken < n  &&  i < iLen ) {
		pToken[nToken] = pStr ;
		iTokenLen[nToken] = 0 ;
		while ( *pStr != ':'  &&  i < iLen ) { 
			pStr ++ ;
			i ++ ;
			iTokenLen[nToken] ++ ;
		}
		if ( *pStr == ':' ) {
			*pStr = 0 ;
			pStr ++ ;
			i ++ ;
		}
		nToken ++ ;
	}

	for ( i = nToken ; i < n ; i ++ ) {
		pToken[i] = NULL ;
		iTokenLen[i] = 0 ;
	}

	return nToken ;

}
/******************************************************************************/
/*  xmemo_SendUdp_Ipmsg_ANSENTRY                                              */
/******************************************************************************/

static int xmemo_SendUdp_Ipmsg_ANSENTRY ( zSocketType *pSocket,
                           zMemberType *pMember, char *pToken [ ], int iTokenLen [ ] )
{


int 	ipmsgcmd ;
char	data [ 100 + MAXLEN_NICKNAME + MAXLEN_USERNAME + MAXLEN_HOSTNAME + MAXLEN_GROUPNAME ] ;
int 	datalen ;
int 	rc ;
struct	sockaddr_in	addr ;

	if ( bMemoTraceLog ) {
		char	sLog [ 64 ] ;
		sprintf ( sLog, "%c://%s:%u",
			pMember->cProtocol,
			ipaddress_long_to_string ( pMember->iAddr ),
			pMember->iPort ) ;
		xmemo_TraceLog ( "xmemo_SendUdp_Ipmsg_ANSENTRY", sLog ) ;
	}

	ipmsgcmd = IPMSG_ANSENTRY ;
	if ( bUserAbsence ) ipmsgcmd |= IPMSG_ABSENCEOPT ;

	sprintf ( data, "%d:%d:%s:%s:%d:",
	                IPMSG_VERSION, 
	                xmemo_IpmsgPacketNo ( ), 
	                sMemoUsername, 
	                sMemoHostname, 
	                ipmsgcmd ) ;

	xmemo_EncodeMessaage_Ipmsg ( &(data[strlen(data)]), sMemoNickname, MAXLEN_NICKNAME ) ;
	datalen = strlen ( data ) + 1 ;
	xmemo_EncodeMessaage_Ipmsg ( &(data[datalen]), sMemoGroupname, MAXLEN_GROUPNAME ) ;
	datalen += strlen ( &(data[datalen]) ) + 1 ;

	memset ( &addr, 0, sizeof(addr) ) ;
	addr.sin_family      = AF_INET ;
	addr.sin_port        = htons ( pMember->iPort ) ;
	addr.sin_addr.s_addr = htonl ( pMember->iAddr ) ;

	rc = sendto ( pSocket->iSock, data, datalen, 0, (struct sockaddr *)&addr, sizeof(addr) ) ;
	if ( rc >= 0 ) {
		rc = RC_SUCCESS ;
	} else {
		rc = RC_ERROR ;
	}
	if ( bMemoTracePacket ) xmemo_TracePacket ( data, datalen ) ;

	return rc ;

}

/******************************************************************************/
/*  xmemo_SendUdp_Ipmsg_RECVMSG                                               */
/******************************************************************************/

static int xmemo_SendUdp_Ipmsg_RECVMSG ( zSocketType *pSocket,
                           zMemberType *pMember, char *pToken [ ], int iTokenLen [ ] )
{

int 	ipmsgcmd ;
char	data [ 100 + MAXLEN_USERNAME + MAXLEN_HOSTNAME ] ;
int 	rc ;
struct	sockaddr_in	addr ;

	if ( bMemoTraceLog ) {
		char	sLog [ 64 ] ;
		sprintf ( sLog, "%c://%s:%u",
			pMember->cProtocol,
			ipaddress_long_to_string ( pMember->iAddr ),
			pMember->iPort ) ;
		xmemo_TraceLog ( "xmemo_SendUdp_Ipmsg_RECVMSG", sLog ) ;
	}

	sprintf ( data, "%d:%d:%s:%s:%d:%s",
	                IPMSG_VERSION, 
	                xmemo_IpmsgPacketNo ( ), 
	                sMemoUsername, 
	                sMemoHostname, 
	                IPMSG_RECVMSG,
	                pToken[1] ) ; 

	memset ( &addr, 0, sizeof(addr) ) ;
	addr.sin_family      = AF_INET ;
	addr.sin_port        = htons ( pMember->iPort ) ;
	addr.sin_addr.s_addr = htonl ( pMember->iAddr ) ;

	rc = sendto ( pSocket->iSock, data, strlen(data)+1, 0, (struct sockaddr *)&addr, sizeof(addr) ) ;
	if ( rc >= 0 ) {
		rc = RC_SUCCESS ;
	} else {
		rc = RC_ERROR ;
	}
	if ( bMemoTracePacket ) xmemo_TracePacket ( data, strlen(data)+1 ) ;

}

/******************************************************************************/
/*  xmemo_RecvUdp_Ipmsg_BR_ENTRY                                              */
/******************************************************************************/

static void xmemo_RecvUdp_Ipmsg_BR_ENTRY ( zSocketType *pSocket,
                           zMemberType *pMember, char *pToken [ ], int iTokenLen [ ] )
{

int 	ipmsgoption ;
char	sNickname [ MAXLEN_NICKNAME ] ;
char	sGroupname [ MAXLEN_GROUPNAME ] ;

	ipmsgoption = atoi ( pToken[4] ) ;

	if ( ( ipmsgoption & IPMSG_ABSENCEOPT ) == 0 ) pMember->bAbsence = FALSE ;
	else                                           pMember->bAbsence = TRUE ;

	if ( pToken[5] != NULL  &&  iTokenLen[5] > 0 ) {
		xmemo_DecodeMessage_Ipmsg ( sNickname, pToken[5], MAXLEN_NICKNAME ) ;
		pMember->pNick = sNickname ;
		if ( strlen ( pToken[5] ) + 1 < iTokenLen[5] ) {
			xmemo_DecodeMessage_Ipmsg ( sGroupname, pToken[5]+strlen(pToken[5])+1, MAXLEN_GROUPNAME ) ;
			pMember->pGroup = sGroupname ;
		}
	}

	xmemo_CreateMember ( pMember ) ;

	xmemo_SendUdp_Ipmsg_ANSENTRY ( pSocket, pMember, pToken, iTokenLen ) ;

}

/******************************************************************************/
/*  xmemo_RecvUdp_Ipmsg_ANSENTRY                                              */
/******************************************************************************/

static void xmemo_RecvUdp_Ipmsg_ANSENTRY ( zSocketType *pSocket,
                           zMemberType *pMember, char *pToken [ ], int iTokenLen [ ] )
{

int 	ipmsgoption ;
char	sNickname [ MAXLEN_NICKNAME ] ;
char	sGroupname [ MAXLEN_GROUPNAME ] ;

	ipmsgoption = atoi ( pToken[4] ) ;

	if ( ( ipmsgoption & IPMSG_ABSENCEOPT ) == 0 ) pMember->bAbsence = FALSE ;
	else                                           pMember->bAbsence = TRUE ;

	if ( pToken[5] != NULL  &&  iTokenLen[5] > 0 ) {
		xmemo_DecodeMessage_Ipmsg ( sNickname, pToken[5], MAXLEN_NICKNAME ) ;
		pMember->pNick = sNickname ;
		if ( strlen ( pToken[5] ) + 1 < iTokenLen[5] ) {
			xmemo_DecodeMessage_Ipmsg ( sGroupname, pToken[5]+strlen(pToken[5])+1, MAXLEN_GROUPNAME ) ;
			pMember->pGroup = sGroupname ;
		}
	}

	xmemo_CreateMember ( pMember ) ;

}

/******************************************************************************/
/*  xmemo_RecvUdp_Ipmsg_SENDMSG                                               */
/******************************************************************************/

static void xmemo_RecvUdp_Ipmsg_SENDMSG ( zSocketType *pSocket,
                           zMemberType *pMember, char *pToken [ ], int iTokenLen [ ] )
{

int 	ipmsgpacket, ipmsgoption ;
time_t	tNow ;
zChainType	*pChain ;
zMemoType	*pMemo ;
char	sText [ MAXLEN_MESSAGE ] ;

zMemoType	zAbsentMemo ;

	time ( &tNow ) ;
	ipmsgpacket = atoi ( pToken[1] ) ;
	ipmsgoption = atoi ( pToken[4] ) ;

	if ( ( ipmsgoption & IPMSG_SENDCHECKOPT ) != 0 ) {
		xmemo_SendUdp_Ipmsg_RECVMSG ( pSocket, pMember, pToken, iTokenLen ) ;
	}

	/* Check whether the received packet is a retry packet */

	for ( pChain = zMemoAnchor.pHead ; pChain != NULL ; pChain = pChain->pNext ) {
		pMemo = (zMemoType*)(pChain->pData) ;
		if ( pMemo->iAddr == pMember->iAddr
		  && pMemo->iPort == pMember->iPort
		  && pMemo->tCreate >= tNow - 60
		  && pMemo->iRecvNo == ipmsgpacket ) {
		  	break ;
		}
	}

	/* If the received packet is a retry packet, then ignore it and return. */

	if ( pChain != NULL ) return ;

	/* Put the received memo onto the desktop screen. */

	pMemo = xmemo_CreateMemo ( ) ;
	if ( pMemo == NULL ) return ;

	pMemo->iAddr = pMember->iAddr ;
	pMemo->iPort = pMember->iPort ;
	pMemo->iRecvNo = ipmsgpacket ;
	if ( pToken[2] != NULL  && iTokenLen[2] > 0 ) {
		pMemo->pUser = XtMalloc ( iTokenLen[2] + 1 ) ;
		if ( pMemo->pUser == NULL ) {
			xmemo_ErrorLog ( "xmemo_RecvUdp_Ipmsg_SENDMSG", "XtMalloc", NULL ) ;
			xmemo_FreeMemo ( pMemo ) ;
			return ;
		}
		strcpy ( pMemo->pUser, pToken[2] ) ;
	}
	if ( pToken[3] != NULL  && iTokenLen[3] > 0 ) {
		pMemo->pHost = XtMalloc ( iTokenLen[3] + 1 ) ;
		if ( pMemo->pHost == NULL ) {
			xmemo_ErrorLog ( "xmemo_RecvUdp_Ipmsg_SENDMSG", "XtMalloc", NULL ) ;
			xmemo_FreeMemo ( pMemo ) ;
			return ;
		}
		strcpy ( pMemo->pHost, pToken[3] ) ;
	}
	if ( pToken[5] != NULL  && iTokenLen[5] > 0 ) {
		xmemo_DecodeMessage_Ipmsg ( sText, pToken[5], MAXLEN_MESSAGE ) ;
		pMemo->pText = XtMalloc ( strlen ( sText ) + 1 ) ;
		if ( pMemo->pText == NULL ) {
			xmemo_ErrorLog ( "xmemo_RecvUdp_Ipmsg_SENDMSG", "XtMalloc", NULL ) ;
			xmemo_FreeMemo ( pMemo ) ;
			return ;
		}
		strcpy ( pMemo->pText, sText ) ;
	}

	xmemo_InsertMemo ( pMemo ) ;

}

/******************************************************************************/
/*  xmemo_RecvUdp_Ipmsg_RECVMSG                                               */
/******************************************************************************/

static void xmemo_RecvUdp_Ipmsg_RECVMSG ( zSocketType *pSocket,
                           zMemberType *pMember, char *pToken [ ], int iTokenLen [ ] )
{

int 	ipmsgpacket ;
zChainType	*pChain ;
zMemberType	*pPeer ;
zMemoType	*pMemo ;

	if ( iTokenLen[5] == 0 ) {
		xmemo_ErrorLog ( "xmemo_RecvUdp_Ipmsg_RECVMSG", "Discard packet", NULL ) ;
		return ;
	}

	ipmsgpacket = atoi ( pToken[5] ) ;

	for ( pChain = zSendPeerAnchor.pHead ; pChain != NULL ; pChain = pChain->pNext ) {
		pPeer = (zMemberType*) pChain->pData ;
		if ( pPeer->cProtocol == pMember->cProtocol
		  && pPeer->iAddr     == pMember->iAddr
		  && pPeer->iPort     == pMember->iPort
		  && pPeer->iSendNo   == ipmsgpacket
		  && pPeer->iState    == XMEMO_STATE_ANY_SENT ) {
		  	break ;
		}
	}

	if ( pChain == NULL ) return ;

	pMemo = pPeer->pSendMemo ;
	pMemo->iSendCount -- ;
	if ( pMemo->iSendCount == 0 ) {
		xmemo_RemoveFromChain ( &zSendMemoAnchor, pMemo->pChain ) ;
		xmemo_FreeMemo ( pMemo ) ;
	}

	xmemo_RemoveFromChain ( &zSendPeerAnchor, pPeer->pChain ) ;

	xmemo_FreeMember ( pPeer ) ;

}

/******************************************************************************/
/*  xmemo_TimeoutUdp_Ipmsg                                                    */
/******************************************************************************/

void xmemo_TimeoutUdp_Ipmsg ( zMemberType *pPeer )
{

zMemoType	*pMemo ;

	pMemo = pPeer->pSendMemo ;

	if ( pPeer->iRetry > XMEMO_IPMSG_RETRYOUT ) {

		pMemo->iSendCount -- ;
		if ( pMemo->iSendCount == 0 ) {
			xmemo_RemoveFromChain ( &zSendMemoAnchor, pMemo->pChain ) ;
			xmemo_FreeMemo ( pMemo ) ;
		}

		xmemo_RemoveFromChain ( &zSendPeerAnchor, pPeer->pChain ) ;
		xmemo_FreeMember ( pPeer ) ;

	} else {

		xmemo_NetSend_Ipmsg ( pPeer, pMemo ) ;
	}

	return ;

}

/******************************************************************************/
/*  xmemo_RecvUdp_Ipmsg                                                       */
/******************************************************************************/

void xmemo_RecvUdp_Ipmsg ( zSocketType *pSocket, long iPeerAddr, short iPeerPort, char *data, int datalen )
{

char	sLog [ 200 ], sBuf [ 20 ] ;
char	*pToken [ 6 ] ;
int 	iTokenCount, iTokenLen [ 6 ] ;
zMemberType	zMember ;

	iTokenCount = xmemo_IpmsgToken ( pToken, iTokenLen, 6, data, datalen ) ;

	if ( iTokenCount < 5
	  || ! xmemo_CheckNumericRange ( pToken[0], 1, 1 )
	  || ! xmemo_CheckNumeric ( pToken[1] )
	  || ! xmemo_CheckNumeric ( pToken[4] ) ) {
		xmemo_ErrorLog ( "xmemo_RecvUdp_Ipmsg", "Discard packet", data ) ;
		return ;
	}

	zMember.cProtocol = XMEMO_PROTOCOL_IPMSG ;
	zMember.iAddr = iPeerAddr ;
	zMember.iPort = pSocket->iPort ;
	zMember.pGroup = NULL ;
	zMember.pUser  = pToken[2] ;
	zMember.pHost  = pToken[3] ;
	zMember.pNick  = NULL ;

	switch ( atoi ( pToken[4] ) & 0x000000FF ) {
	case IPMSG_BR_ENTRY   : xmemo_RecvUdp_Ipmsg_BR_ENTRY ( pSocket, &zMember, pToken, iTokenLen ) ; break ;
	case IPMSG_BR_EXIT    : xmemo_DeleteMember           ( &zMember )                  ; break ;
	case IPMSG_ANSENTRY   : xmemo_RecvUdp_Ipmsg_ANSENTRY ( pSocket, &zMember, pToken, iTokenLen ) ; break ;
	case IPMSG_BR_ABSENCE : xmemo_RecvUdp_Ipmsg_ANSENTRY ( pSocket, &zMember, pToken, iTokenLen ) ; break ;
	case IPMSG_SENDMSG    : xmemo_RecvUdp_Ipmsg_SENDMSG  ( pSocket, &zMember, pToken, iTokenLen ) ; break ;
	case IPMSG_RECVMSG    : xmemo_RecvUdp_Ipmsg_RECVMSG  ( pSocket, &zMember, pToken, iTokenLen ) ; break ;
	case IPMSG_READMSG    :
	case IPMSG_DELMSG     :
	default :break ;
	}

}

/******************************************************************************/
/*  xmemo_NetSend_Ipmsg                                                       */
/******************************************************************************/

int xmemo_NetSend_Ipmsg ( zMemberType* pPeer, zMemoType* pMemo )
{

char	data [ 100 + MAXLEN_USERNAME + MAXLEN_HOSTNAME + MAXLEN_MESSAGE ] ;
int 	rc ;
struct	sockaddr_in	addr ;

	if ( bMemoTraceLog ) {
		char	sLog [ 64 ] ;
		sprintf ( sLog, "%c://%s:%u    Retry=%d",
			pPeer->cProtocol,
			ipaddress_long_to_string ( pPeer->iAddr ),
			pPeer->iPort, pPeer->iRetry ) ;
		xmemo_TraceLog ( "xmemo_NetSend_Ipmsg", sLog ) ;
	}

	if ( pPeer->iRetry > XMEMO_IPMSG_RETRYOUT ) {
		pMemo->iSendCount -- ;
		if ( pMemo->iSendCount == 0 ) {
			xmemo_RemoveFromChain ( &zSendMemoAnchor, pMemo->pChain ) ;
			xmemo_FreeMemo ( pMemo ) ;
		}
		xmemo_RemoveFromChain ( &zSendPeerAnchor, pPeer->pChain ) ;
		xmemo_FreeMember ( pPeer ) ;
		return RC_SUCCESS ;
	} else if ( pPeer->iRetry == 0 ) {
		pPeer->iSendNo = xmemo_IpmsgPacketNo ( ) ;
	}

	pPeer->iRetry ++ ;

	sprintf ( data, "%d:%d:%s:%s:%d:",
	                IPMSG_VERSION, 
	                pPeer->iSendNo, 
	                sMemoUsername, 
	                sMemoHostname, 
	                IPMSG_SENDMSG | IPMSG_SENDCHECKOPT ) ;

	xmemo_EncodeMessaage_Ipmsg ( &(data[strlen(data)]), pMemo->pText, MAXLEN_MESSAGE ) ;

	memset ( &addr, 0, sizeof(addr) ) ;
	addr.sin_family      = AF_INET ;
	addr.sin_port        = htons ( pPeer->iPort ) ;
	addr.sin_addr.s_addr = htonl ( pPeer->iAddr ) ;

	rc = sendto ( pPeer->pUdpSocket->iSock, data, strlen(data)+1, 0, (struct sockaddr *)&addr, sizeof(addr) ) ;
	if ( rc >= 0 ) {
		rc = RC_SUCCESS ;
	} else {
		rc = RC_ERROR ;
	}
	if ( bMemoTracePacket ) xmemo_TracePacket ( data, strlen(data)+1 ) ;

	pPeer->iState = XMEMO_STATE_ANY_SENT ;

	/* Start timer */

	time ( &(pPeer->tTimeout) ) ;
	pPeer->tTimeout += XMEMO_IPMSG_TIMEOUT ;
	xmemo_StartTimer ( pPeer->tTimeout ) ;

	return rc ;

}

/******************************************************************************/
/*  xmemo_NetNotify_Ipmsg                                                     */
/******************************************************************************/

int xmemo_NetNotify_Ipmsg ( int iSock, long iAddr, short iPort, int iNotify )
{

int 	ipmsgcmd ;
char	data [ 100 + MAXLEN_NICKNAME + MAXLEN_USERNAME + MAXLEN_HOSTNAME + MAXLEN_GROUPNAME ] ;
int 	datalen ;
int 	rc ;
struct	sockaddr_in	addr ;

	if ( bMemoTraceLog ) {
		char	sLog [ 32 ] ;
		sprintf ( sLog, "%c://%s:%u",
			XMEMO_PROTOCOL_IPMSG,
			ipaddress_long_to_string ( iAddr ),
			(unsigned short) iPort ) ;
		xmemo_TraceLog ( "xmemo_NetNotify_Ipmsg", sLog ) ;
	}

	switch ( iNotify ) {

	case XMEMO_NET_NOTIFY_ENTER :

		ipmsgcmd = IPMSG_BR_ENTRY ;
		if ( bUserAbsence ) ipmsgcmd |= IPMSG_ABSENCEOPT ;
		break;

	case XMEMO_NET_NOTIFY_REPLY :

		ipmsgcmd = IPMSG_ANSENTRY ;
		if ( bUserAbsence ) ipmsgcmd |= IPMSG_ABSENCEOPT ;
		break;

	case XMEMO_NET_NOTIFY_LEAVE :

		ipmsgcmd = IPMSG_BR_EXIT ;
		break;

	case XMEMO_NET_NOTIFY_CHANGE :

		ipmsgcmd = IPMSG_BR_ABSENCE ;
		if ( bUserAbsence ) ipmsgcmd |= IPMSG_ABSENCEOPT ;
		break ;

	case XMEMO_NET_NOTIFY_PRESENCE :

		ipmsgcmd = IPMSG_BR_ABSENCE ;
		break;

	case XMEMO_NET_NOTIFY_ABSENCE :

		ipmsgcmd = IPMSG_BR_ABSENCE + IPMSG_ABSENCEOPT ;
		break;

	}

	sprintf ( data, "%d:%d:%s:%s:%d:",
	                IPMSG_VERSION, 
	                xmemo_IpmsgPacketNo ( ), 
	                sMemoUsername, 
	                sMemoHostname, 
	                ipmsgcmd ) ;

	xmemo_EncodeMessaage_Ipmsg ( &(data[strlen(data)]), sMemoNickname, MAXLEN_NICKNAME ) ;
	datalen = strlen ( data ) + 1 ;
	xmemo_EncodeMessaage_Ipmsg ( &(data[datalen]), sMemoGroupname, MAXLEN_GROUPNAME ) ;
	datalen += strlen ( &(data[datalen]) ) + 1 ;

	memset ( &addr, 0, sizeof(addr) ) ;
	addr.sin_family      = AF_INET ;
	addr.sin_port        = htons ( iPort ) ;
	addr.sin_addr.s_addr = htonl ( iAddr ) ;

	rc = sendto ( iSock, data, datalen, 0, (struct sockaddr *)&addr, sizeof(addr) ) ;
	if ( rc >= 0 ) {
		rc = RC_SUCCESS ;
	} else {
		rc = RC_ERROR ;
	}
	if ( bMemoTracePacket ) xmemo_TracePacket ( data, datalen ) ;

	return rc ;

}
