/******************************************************************************/
/*                                                                            */
/*  <<< Copyright >>>                                                         */
/*                                                                            */
/*  The "xmemo" software is copyright 1999 by Takao Abe.                      */
/*  All rights reserved.                                                      */
/*                                                                            */
/*  The "xmemo" software refers to the source code, the binary code,          */
/*  the documents, and all derivatives based on this software under           */
/*  the copyright law, that is a software containing a program or a           */
/*  portion of a program, regardless of modifications and/or translations.    */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/*  <<< Licence and redistribution >>>                                        */
/*                                                                            */
/*  The source code and the binary code are free software.                    */
/*  The use and redistribution of source code and binary code are permitted   */
/*  provided that the following conditions are met:                           */
/*  You may use the software for private use and commercial use,              */
/*  and may modify your copy or copies of the software or any portion of it.  */
/*                                                                            */
/*  THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESSED OR IMPLIED            */
/*  WARRANTIES OF ANY KIND, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED        */
/*  WARRANTIES OF MERCHANTABLILITY AND FITNESS FOR A PARTICULAR PURPOSE.      */
/*  IN NO EVENT SHALL THE AUTHOR TAKAO ABE BE LIABLE FOR ANY DIRECT,          */
/*  INDIRECT, GENERAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES           */
/*  ( INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR       */
/*  SERVICES; LOSS OF USE, DATA OR PROFITS; OR BUSINESS INTERRUPTION )        */
/*  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,       */
/*  STRICT LIABILITY, OR TORT ( INCLUDING NEGLIGENCE OR OTHERWISE )           */
/*  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED       */
/*  OF THE POSSIBILITY OF SUCH DAMAGE.                                        */
/*                                                                            */
/*  Redistribution and use in source form or binary form, with or without     */
/*  modification, the whole program or a portion of it, are permitted         */
/*  provided that one retains the entire copyright notice properly, and       */
/*  displays or retains this license in the documentation and/or other        */
/*  materials provided with the distribution.                                 */
/*                                                                            */
/*  Redistribution and use in source form or binary form, with modification,  */
/*  the whole program or a portion of it, are permitted provided that one     */
/*  declares that it is modified software and not original, and displays or   */
/*  describes any modifications in the documentation and/or other materials   */
/*  provided with the distribution.                                           */
/*                                                                            */
/*  This software and the name of the author must not be used to endorse or   */
/*  promote products derived from this software without prior written         */
/*  permission.                                                               */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/*  EMail   abetakao@bea.hi-ho.ne.jp                                          */
/*          HCA00763@nifty.ne.jp                                              */
/*                                                                            */
/*  URL     http://www.bea.hi-ho.ne.jp/abetakao/                              */
/*                                                                            */
/******************************************************************************/


#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<time.h>
#include	<errno.h>
#include	<sys/types.h>
#include	<sys/time.h>
#include	<sys/ioctl.h>
#include	<sys/socket.h>
#ifdef __sparc
#include	<sys/filio.h>
#endif
#include	<netdb.h>
#include	<netinet/in.h>
#include	<arpa/inet.h>


#include	<X11/Intrinsic.h>


#include	"xmemo.h"
#include	"xmemodata.h"
#include	"xmemofunc.h"


/******************************************************************************/
/*  Application protocol and format : Oboegaki                                */
/******************************************************************************/

#define		OBOEGAKI_PACKET_HELLO		"Hi,Oboe!"
#define		OBOEGAKI_PACKET_ACK  		0

#define		OBOEGAKI_STRUCT_STRING		"OboeData"
#define		OBOEGAKI_STRUCT_VERSION		0x0100
#define		OBOEGAKI_VERSION_NUMBER		0
#define		OBOEGAKI_VERSION_STRING		"2.00"

#define		OBOEGAKI_NOTIFY_ENTER		1
#define		OBOEGAKI_NOTIFY_REPLY		2
#define		OBOEGAKI_NOTIFY_LEAVE		3

struct OboegakiEntryStruct {
	int 	iPacketSize ;
	char	sPacketString [ 8 ] ;	/* "OboeBct"  */
	int 	iPluginVersion ;		/* 0x00000100 */
	int 	iNotify         ;		/* OBOEGAKI_NOTIFY_XXXXX */
	int 	bAbsence        ;		/* 1:Absence  0:Presence */
	int 	iOffsetNickname ;
	int 	iOffsetGroupname ;
	int 	iOffsetUsername ;
	int 	iOffsetHostname ;
} ;

struct OboegakiDataStruct {
/*	char	sHeader[9] ;    */		/* OBOEGAKI_STRUCT_STRING */
	int 	iStructSize ;   		/* sizeof(struct OboegakiDataStruct)+Message */
	int 	iStructVersion ;		/* OBOEGAKI_STRUCT_VERSION */
	int 	iVersion ;      		/* OBOEGAKI_VERSION_NUMBER */
	int 	iOffsetVersion ;		/* Offset to OBOEGAKI_VERSION_STRING */
	int 	iOffsetTitle ;  		/* Offset to XMEMO_OBOEGAKI_TITLE */
	int 	iOffsetData ;   		/* Offset to Data */

	int 	iOffsetFont ;   		/*   */
	int 	iForeColor ;    		/* Microsoft Windows COLORREF struct */
	int 	iBackColor ;    		/* Microsoft Windows COLORREF struct */
	int 	iIcon ;         		/* 0 */

	int 	bShellExec ;    		/* FALSE */
	int 	iOffsetProgName ;		/* 0 */
	int 	iOffsetFileName ;		/* 0 */

	int 	bTimerEnable ;   		/* FALSE */
	int 	iTimerFiller[10] ;		/* 0's */

	int 	bSendMaster ;   		/* TRUE */
	int 	iPositionSignature ;	/* 1 */
	int 	iOffsetSignature ;		/* Offset to user@host */
	int 	bSendTime ;     		/* FALSE */
	int 	bSendResize ;   		/* TRUE */
	int 	iRectLeft, iRectTop, iRectRight, iRectBottom ;
	int 	bSendCenter ;   		/* TRUE */
	int 	bSendOptions[4] ;		/* 0's */
	int 	iOffsetFrom ;   		/* Offset to From */
	int 	iOffsetTo ;     		/* Offset to To */

	int 	bVoteYesNo ;    		/* FALSE */
	int 	bVoteOther ;    		/* FALSE */
	int 	iVoteFiller[5] ;		/* 0's */

	int 	iCreateYear, iCreateMonth, iCreateDay ;
	int 	iValidTerm ;

	int 	iDxlId ;
	int 	iDxlFiller[3] ;
} ;

/******************************************************************************/
/*  Local constant definition                                                 */
/******************************************************************************/

#define		XMEMO_OBOEGAKI_TITLE		"xmemo"
#define		XMEMO_OBOEGAKI_TIMEOUT		5

/******************************************************************************/
/*  xmemo_DecodeMessage_Oboegaki                                              */
/******************************************************************************/

static int xmemo_DecodeMessage_Oboegaki ( char *pDest, char *pSource, int iLen )
{

static	char	sCR [ ] = { 0x00, 0x00, 0x00 } ;
static	char	sLF [ ] = { 0x0A, 0x00, 0x00 } ;
int 	rc ;

	switch ( cMemoCharset ) {
	case CHARSET_EUC :
		rc = ShiftJIS_To_EUC ( pDest, pSource, iLen, 0, 0x0A ) ;
		break ;
	case CHARSET_JIS7 :
		rc = ShiftJIS_To_JIS7 ( pDest, pSource, iLen, 0, 0x0A ) ;
		break ;
	case CHARSET_JIS8 :
		rc = ShiftJIS_To_JIS8 ( pDest, pSource, iLen, 0, 0x0A ) ;
		break ;
	default :
		rc = ShiftJIS_CRLF ( pDest, pSource, iLen, sCR, sLF ) ;
		break ;
	}

	return rc ;

}

/******************************************************************************/
/*  xmemo_EncodeMessage_Oboegaki                                              */
/******************************************************************************/

static int xmemo_EncodeMessage_Oboegaki ( char *pDest, char *pSource, int iLen )
{

static	char	sCR [ ] = { 0x00, 0x00, 0x00 } ;
static	char	sLF [ ] = { 0x0D, 0x0A, 0x00 } ;
int 	rc ;

	if ( pSource == NULL ) {
		*pDest = 0 ;
		return 0 ;
	}

	switch ( cMemoCharset ) {
	case CHARSET_EUC :
		rc = ISO_To_ShiftJIS ( pDest, pSource, iLen, ISO_DefaultDesignation_EUC( ), sLF ) ;
		break ;
	case CHARSET_JIS7 :
	case CHARSET_JIS8 :
		rc = ISO_To_ShiftJIS ( pDest, pSource, iLen, ISO_DefaultDesignation_JIS( ), sLF ) ;
		break ;
	default :
		rc = ShiftJIS_CRLF ( pDest, pSource, iLen, sCR, sLF ) ;
		break ;
	}

	return rc ;

}

/******************************************************************************/
/*  xmemo_RecvUdp_Oboegaki                                                    */
/******************************************************************************/

void xmemo_RecvUdp_Oboegaki ( zSocketType *pSocket, long iPeerAddr, short iPeerPort, char *data, int datalen )
{

char	sLog [ 200 ], sBuf [ 20 ] ;
struct	OboegakiEntryStruct	zOboegakiHeader ;

zMemberType	zMember ;
char	sNickname [ MAXLEN_NICKNAME ] ;
char	sGroupname [ MAXLEN_GROUPNAME ] ;

	if ( datalen < sizeof(struct OboegakiEntryStruct) ) {
		sprintf ( sLog, "Packet size = %d", datalen ) ;
		xmemo_ErrorLog ( "xmemo_RecvUdp_Oboegaki", "Discard packet", sLog ) ;
		return ;
	}

	memcpy ( (char*)&zOboegakiHeader, data,  sizeof(zOboegakiHeader) ) ;

	zOboegakiHeader.iPacketSize      = int_from_lobyte_first ( zOboegakiHeader.iPacketSize      ) ;
	zOboegakiHeader.iPluginVersion   = int_from_lobyte_first ( zOboegakiHeader.iPluginVersion   ) ;
	zOboegakiHeader.iNotify          = int_from_lobyte_first ( zOboegakiHeader.iNotify          ) ;
	zOboegakiHeader.bAbsence         = int_from_lobyte_first ( zOboegakiHeader.bAbsence         ) ;
	zOboegakiHeader.iOffsetNickname  = int_from_lobyte_first ( zOboegakiHeader.iOffsetNickname  ) ;
	zOboegakiHeader.iOffsetGroupname = int_from_lobyte_first ( zOboegakiHeader.iOffsetGroupname ) ;
	zOboegakiHeader.iOffsetUsername  = int_from_lobyte_first ( zOboegakiHeader.iOffsetUsername  ) ;
	zOboegakiHeader.iOffsetHostname  = int_from_lobyte_first ( zOboegakiHeader.iOffsetHostname  ) ;
	
	if ( zOboegakiHeader.iPacketSize > datalen
	  || strncmp ( zOboegakiHeader.sPacketString, "OboeBct", 8 ) != 0
	  || zOboegakiHeader.iPluginVersion   != 0x00000100
	  || zOboegakiHeader.iOffsetNickname  >= zOboegakiHeader.iPacketSize
	  || zOboegakiHeader.iOffsetGroupname >= zOboegakiHeader.iPacketSize
	  || zOboegakiHeader.iOffsetUsername  >= zOboegakiHeader.iPacketSize
	  || zOboegakiHeader.iOffsetHostname  >= zOboegakiHeader.iPacketSize ) {
		strncpy ( sBuf, zOboegakiHeader.sPacketString, 9 ) ;
		sBuf[8] = 0 ;
		sprintf ( sLog, "%d  %s  %08X  %d  %d  %d  %d  %d  %d", 
				zOboegakiHeader.iPacketSize, sBuf,
				zOboegakiHeader.iPluginVersion  ,
				zOboegakiHeader.iNotify         ,
				zOboegakiHeader.bAbsence        ,
 	  			zOboegakiHeader.iOffsetNickname ,
				zOboegakiHeader.iOffsetGroupname,
				zOboegakiHeader.iOffsetUsername ,
				zOboegakiHeader.iOffsetHostname ) ;
		xmemo_ErrorLog ( "xmemo_RecvUdp_Oboegaki", "Discard packet", sLog ) ;
		return ;
	}

	zMember.cProtocol = XMEMO_PROTOCOL_OBOEGAKI ;
	zMember.iAddr = iPeerAddr ;
	zMember.iPort = pSocket->iPort ;
	zMember.pGroup = zMember.pNick = zMember.pUser = zMember.pHost = NULL ;
	if ( zOboegakiHeader.iOffsetNickname  >= sizeof(zOboegakiHeader) ) {
		xmemo_DecodeMessage_Oboegaki ( sNickname, data + zOboegakiHeader.iOffsetNickname, MAXLEN_NICKNAME ) ;
		zMember.pNick  = sNickname ;
	}
	if ( zOboegakiHeader.iOffsetGroupname >= sizeof(zOboegakiHeader) ) {
		xmemo_DecodeMessage_Oboegaki ( sGroupname, data + zOboegakiHeader.iOffsetGroupname, MAXLEN_GROUPNAME ) ;
		zMember.pGroup = sGroupname ;
	}
	if ( zOboegakiHeader.iOffsetUsername  >= sizeof(zOboegakiHeader) ) {
		zMember.pUser  = data + zOboegakiHeader.iOffsetUsername ;
	}
	if ( zOboegakiHeader.iOffsetHostname  >= sizeof(zOboegakiHeader) ) {
		zMember.pHost  = data + zOboegakiHeader.iOffsetHostname ;
	}
	if ( zOboegakiHeader.bAbsence == 0 ) zMember.bAbsence = FALSE ;
	else                                 zMember.bAbsence = TRUE ;

	switch ( zOboegakiHeader.iNotify ) {
	case OBOEGAKI_NOTIFY_ENTER :
		xmemo_CreateMember ( &zMember ) ;
		xmemo_NetNotify_Oboegaki ( pSocket->iSock, zMember.iAddr, zMember.iPort, XMEMO_NET_NOTIFY_REPLY ) ;
		break ;
	case OBOEGAKI_NOTIFY_REPLY :
		xmemo_CreateMember ( &zMember ) ;
		break ;
	case OBOEGAKI_NOTIFY_LEAVE :
		xmemo_DeleteMember ( &zMember ) ;
		break ;
	}

}

/******************************************************************************/
/*  xmemo_SendConnectedTcp_Oboegaki                                           */
/******************************************************************************/

static int xmemo_SendConnectedTcp_Oboegaki ( zMemberType *pMember )
{

int 	i, rc ;
struct	OboegakiDataStruct	zDataHdr ;
char	data [ MAXLEN_PACKET ] ;
int 	iOffset ;
int 	iColor ;
zSocketType	*pSocket ;
zMemoType	*pMemo ;

	if ( bMemoTraceLog ) {
		char	sLog [ 64 ] ;
		sprintf ( sLog, "%c://%s:%u    Retry=%d",
			pMember->cProtocol,
			ipaddress_long_to_string ( pMember->iAddr ),
			pMember->iPort, pMember->iRetry ) ;
		xmemo_TraceLog ( "xmemo_SendConnectedTcp_Oboegaki", sLog ) ;
	}

	pSocket = pMember->pTcpSocket ;
	pMemo = pMember->pSendMemo ;

	iOffset = sizeof(int) + strlen(OBOEGAKI_STRUCT_STRING) + 1 + sizeof(struct OboegakiDataStruct) ;

	memset ( &(zDataHdr), 0, sizeof(zDataHdr) ) ;
	zDataHdr.iStructVersion = int_to_lobyte_first ( OBOEGAKI_STRUCT_VERSION ) ;
	zDataHdr.iVersion       = int_to_lobyte_first ( OBOEGAKI_VERSION_NUMBER ) ;
	zDataHdr.iOffsetVersion = int_to_lobyte_first ( iOffset - sizeof(int) ) ;
	strcpy ( &data[iOffset], OBOEGAKI_VERSION_STRING ) ;
	iOffset += strlen ( OBOEGAKI_VERSION_STRING ) + 1 ;
	zDataHdr.iOffsetTitle   = int_to_lobyte_first ( iOffset - sizeof(int) ) ;
	strcpy ( &data[iOffset], XMEMO_OBOEGAKI_TITLE ) ;
	iOffset += strlen ( XMEMO_OBOEGAKI_TITLE ) + 1 ;

	zDataHdr.iOffsetFont = 0xFFFFFFFF ;
	iColor = ( pMemo->iFontColor[0] >> 8 ) & 0x000000FF  	/* Red   */
	       | ( pMemo->iFontColor[1]      ) & 0x0000FF00  	/* Green */
	       | ( pMemo->iFontColor[2] << 8 ) & 0x00FF0000 ;	/* Blue  */
	zDataHdr.iForeColor  = int_to_lobyte_first ( iColor ) ; /* Byte order : RGB */
	iColor = ( pMemo->iBackColor[0] >> 8 ) & 0x000000FF  	/* Red   */
	       | ( pMemo->iBackColor[1]      ) & 0x0000FF00  	/* Green */
	       | ( pMemo->iBackColor[2] << 8 ) & 0x00FF0000 ;	/* Blue  */
	zDataHdr.iBackColor  = int_to_lobyte_first ( iColor ) ; /* Byte order : RGB */
	zDataHdr.bSendMaster = TRUE ;
	zDataHdr.iPositionSignature = 1 ;
	zDataHdr.bSendResize = TRUE ;
	zDataHdr.iRectLeft   = int_to_lobyte_first ( pMemo->x ) ;
	zDataHdr.iRectTop    = int_to_lobyte_first ( pMemo->y ) ;
	zDataHdr.iRectRight  = int_to_lobyte_first ( pMemo->x + pMemo->w - 1 ) ;
	zDataHdr.iRectBottom = int_to_lobyte_first ( pMemo->y + pMemo->h - 1 ) ;
	zDataHdr.bSendCenter = FALSE ;
	zDataHdr.iOffsetSignature = int_to_lobyte_first ( iOffset - sizeof(int) ) ;
	zDataHdr.iOffsetFrom = int_to_lobyte_first ( iOffset - sizeof(int) ) ;
	strcpy ( &data[iOffset], sMemoUsername ) ;
	iOffset += strlen ( sMemoUsername ) ;
	strcpy ( &data[iOffset], "@" ) ;
	iOffset += 1 ;
	strcpy ( &data[iOffset], sMemoHostname ) ;
	iOffset += strlen ( sMemoHostname ) + 1 ;

	zDataHdr.iOffsetData = int_to_lobyte_first ( iOffset - sizeof(int) ) ;
	i = xmemo_EncodeMessage_Oboegaki ( &data[iOffset], pMemo->pText, sizeof(data)-iOffset-2 ) ;
	iOffset += i + 1 ;

	zDataHdr.iStructSize = int_to_lobyte_first ( iOffset - sizeof(int) ) ;

	strcpy ( &data[sizeof(int)], OBOEGAKI_STRUCT_STRING ) ;
	memcpy ( &data[sizeof(int)+strlen(OBOEGAKI_STRUCT_STRING)+1], &zDataHdr, sizeof(struct OboegakiDataStruct) ) ;

	i = int_to_lobyte_first ( iOffset ) ;
	memcpy ( data, &i, sizeof(int) ) ;

	rc = send ( pSocket->iSock, data, iOffset, 0 ) ;
	if ( bMemoTracePacket ) xmemo_TracePacket ( data, iOffset ) ;

	pMember->iState = XMEMO_STATE_ANY_SENT ;

	/* Start timer */

	time ( &(pMember->tTimeout) ) ;
	pMember->tTimeout += XMEMO_OBOEGAKI_TIMEOUT ;
	xmemo_StartTimer ( pMember->tTimeout ) ;

	return RC_SUCCESS ;

}

/******************************************************************************/
/*  xmemo_SendAcceptedTcp_Oboegaki_Hello                                      */
/******************************************************************************/

static int xmemo_SendAcceptedTcp_Oboegaki_Hello ( int iClientSock )
{

int 	rc ;

	if ( bMemoTraceLog ) xmemo_TraceLog ( "xmemo_SendAcceptedTcp_Oboegaki_Hello", NULL ) ;

	rc = send ( iClientSock, OBOEGAKI_PACKET_HELLO, strlen ( OBOEGAKI_PACKET_HELLO ) + 1, 0 ) ;
	if ( rc != strlen ( OBOEGAKI_PACKET_HELLO ) + 1 ) {
		xmemo_ErrorLog ( "xmemo_SendAcceptedTcp_Oboegaki_Hello", "send", NULL ) ;
		return RC_ERROR ;
	}
	if ( bMemoTracePacket ) xmemo_TracePacket ( OBOEGAKI_PACKET_HELLO, strlen ( OBOEGAKI_PACKET_HELLO ) + 1 ) ;

	return RC_SUCCESS ;

}

/******************************************************************************/
/*  xmemo_SendAcceptedTcp_Oboegaki_Ack                                        */
/******************************************************************************/

static int xmemo_SendAcceptedTcp_Oboegaki_Ack ( int iClientSock )
{

int 	iAck ;
int 	rc ;

	if ( bMemoTraceLog ) xmemo_TraceLog ( "xmemo_SendAcceptedTcp_Oboegaki_Ack", NULL ) ;

	iAck = int_to_lobyte_first ( 0 ) ;
	rc = send ( iClientSock, (char*)&iAck, sizeof(int), 0 ) ;
	if ( rc != sizeof(int) ) {
		xmemo_ErrorLog ( "xmemo_SendAcceptedTcp_Oboegaki_Ack", "send", NULL ) ;
		return RC_ERROR ;
	}
	if ( bMemoTracePacket ) xmemo_TracePacket ( (char*)&iAck, sizeof(int) ) ;

	return RC_SUCCESS ;

}

/******************************************************************************/
/*  xmemo_TimeoutConnectedTcp_Oboegaki                                        */
/******************************************************************************/

void xmemo_TimeoutConnectedTcp_Oboegaki ( zMemberType *pPeer )
{

zMemoType	*pMemo ;

	if ( bMemoTraceLog ) {
		char	sLog [ 64 ] ;
		sprintf ( sLog, "%c://%s:%u",
			pPeer->cProtocol,
			ipaddress_long_to_string ( pPeer->iAddr ),
			pPeer->iPort ) ;
		xmemo_TraceLog ( "xmemo_TimeoutConnectedTcp_Oboegaki", sLog ) ;
	}

	pMemo = pPeer->pSendMemo ;
	pMemo->iSendCount -- ;
	if ( pMemo->iSendCount == 0 ) {
		xmemo_RemoveFromChain ( &zSendMemoAnchor, pMemo->pChain ) ;
		xmemo_FreeMemo ( pMemo ) ;
	}

	xmemo_DeleteTcpSocket ( pPeer->pTcpSocket ) ;

	xmemo_RemoveFromChain ( &zSendPeerAnchor, pPeer->pChain ) ;
	xmemo_FreeMember ( pPeer ) ;

	return ;

}

/******************************************************************************/
/*  xmemo_RecvConnectedTcp_Oboegaki                                           */
/******************************************************************************/

void xmemo_RecvConnectedTcp_Oboegaki ( zMemberType *pPeer, char *data, int datalen )
{

zMemoType	*pMemo ;

	if ( datalen < 0 ) {

		xmemo_ErrorLog ( "xmemo_RecvConnectedTcp_Oboegaki", "recv", NULL ) ;

		pMemo = pPeer->pSendMemo ;
		pMemo->iSendCount -- ;
		if ( pMemo->iSendCount == 0 ) {
			xmemo_RemoveFromChain ( &zSendMemoAnchor, pMemo->pChain ) ;
			xmemo_FreeMemo ( pMemo ) ;
		}

		xmemo_DeleteTcpSocket ( pPeer->pTcpSocket ) ;

		xmemo_RemoveFromChain ( &zSendPeerAnchor, pPeer->pChain ) ;
		xmemo_FreeMember ( pPeer ) ;

	} else if ( datalen == 0 ) {

		pMemo = pPeer->pSendMemo ;
		pMemo->iSendCount -- ;
		if ( pMemo->iSendCount == 0 ) {
			xmemo_RemoveFromChain ( &zSendMemoAnchor, pMemo->pChain ) ;
			xmemo_FreeMemo ( pMemo ) ;
		}

		xmemo_DeleteTcpSocket ( pPeer->pTcpSocket ) ;

		xmemo_RemoveFromChain ( &zSendPeerAnchor, pPeer->pChain ) ;
		xmemo_FreeMember ( pPeer ) ;

	} else if ( datalen == sizeof(int) ) {
		/* Received an acknowledge response from the peer */

		pMemo = pPeer->pSendMemo ;
		pMemo->iSendCount -- ;
		if ( pMemo->iSendCount == 0 ) {
			xmemo_RemoveFromChain ( &zSendMemoAnchor, pMemo->pChain ) ;
			xmemo_FreeMemo ( pMemo ) ;
		}

		xmemo_DeleteTcpSocket ( pPeer->pTcpSocket ) ;

		xmemo_RemoveFromChain ( &zSendPeerAnchor, pPeer->pChain ) ;
		xmemo_FreeMember ( pPeer ) ;

	} else if ( datalen == strlen(OBOEGAKI_PACKET_HELLO) + 1  &&  strcmp ( data, OBOEGAKI_PACKET_HELLO ) == 0 ) {
		/* Received a message from the peer */

		xmemo_SendConnectedTcp_Oboegaki ( pPeer ) ;

	} else {

		xmemo_ErrorLog ( "xmemo_RecvConnectedTcp_Oboegaki", "Discard", NULL ) ;

	}

	return ;

}

/******************************************************************************/
/*  xmemo_TimeoutAcceptedTcp_Oboegaki                                         */
/******************************************************************************/

void xmemo_TimeoutAcceptedTcp_Oboegaki ( zSocketType *pSocket )
{

	if ( bMemoTraceLog ) {
		char	sLog [ 64 ] ;
		sprintf ( sLog, "%c://%s:%u",
			pSocket->cProtocol,
			ipaddress_long_to_string ( pSocket->iAddr ),
			pSocket->iPort ) ;
		xmemo_TraceLog ( "xmemo_TimeoutAcceptedTcp_Oboegaki", sLog ) ;
	}

	xmemo_DeleteTcpSocket ( pSocket ) ;

	return ;

}

/******************************************************************************/
/*  xmemo_RecvAcceptedTcp_Oboegaki                                            */
/******************************************************************************/

static void xmemo_RecvAcceptedTcp_Oboegaki ( caddr_t pClientData, int *pSock, XtInputId *pId )
{

int 	rc ;
char	data [ MAXLEN_PACKET + 4 ] ;
int 	iOffsetMin, iOffsetMax ;
struct	OboegakiDataStruct	hdr ;
char	sMessage [ MAXLEN_MESSAGE + MAXLEN_NICKNAME ] ;
int 	iMsgLen ;
zMemoType	*pMemo ;
zSocketType	*pSocket ;

zMemoType	zAbsentMemo ;

	pSocket = (zSocketType*) pClientData ;

	if ( bMemoTraceLog ) {
		char	sLog [ 32 ] ;
		sprintf ( sLog, "%c://%s:%u",
			pSocket->cProtocol,
			ipaddress_long_to_string ( pSocket->iAddr ),
			pSocket->iPort ) ;
		xmemo_TraceLog ( "xmemo_RecvAcceptedTcp_Oboegaki", sLog ) ;
	}

	rc = recv ( pSocket->iSock, data, MAXLEN_PACKET, 0 ) ;
	if ( rc < 0 ) {
		xmemo_ErrorLog ( "xmemo_RecvAcceptedTcp_Oboegaki", "recv", NULL ) ;
		xmemo_DeleteTcpSocket ( pSocket ) ;
		return ;
	} else if ( rc == 0 ) {
		xmemo_DeleteTcpSocket ( pSocket ) ;
		return ;
	} else if ( rc < sizeof(int) + strlen(OBOEGAKI_STRUCT_STRING) + 1 + sizeof(struct OboegakiDataStruct) ) {
		xmemo_ErrorLog ( "xmemo_RecvAcceptedTcp_Oboegaki", "Discard", NULL ) ;
		return ;
	}
	data[rc] = 0 ;
	if ( bMemoTracePacket ) xmemo_TracePacket ( data, rc ) ;

	/* Send an acknowledge response to the peer */

	rc = xmemo_SendAcceptedTcp_Oboegaki_Ack ( pSocket->iSock ) ;

	/* Create memo */

	memcpy ( &hdr, &(data[sizeof(int)+strlen(OBOEGAKI_STRUCT_STRING)+1]), sizeof(struct OboegakiDataStruct) ) ;

	hdr.iStructSize        = int_from_lobyte_first ( hdr.iStructSize ) ;
	hdr.iStructVersion     = int_from_lobyte_first ( hdr.iStructVersion ) ;
	hdr.iVersion           = int_from_lobyte_first ( hdr.iVersion ) ;
	hdr.iOffsetVersion     = int_from_lobyte_first ( hdr.iOffsetVersion ) ;
	hdr.iOffsetTitle       = int_from_lobyte_first ( hdr.iOffsetTitle ) ;
	hdr.iOffsetData        = int_from_lobyte_first ( hdr.iOffsetData ) ;

	hdr.iOffsetFont        = int_from_lobyte_first ( hdr.iOffsetFont ) ;
	hdr.iForeColor         = int_from_lobyte_first ( hdr.iForeColor ) ;  /* Byte order : RGB */
	hdr.iBackColor         = int_from_lobyte_first ( hdr.iBackColor ) ;  /* Byte order : RGB */
	hdr.iIcon              = int_from_lobyte_first ( hdr.iIcon ) ;

	hdr.bSendMaster        = int_from_lobyte_first ( hdr.bSendMaster ) ;
	hdr.iPositionSignature = int_from_lobyte_first ( hdr.iPositionSignature ) ;
	hdr.iOffsetSignature   = int_from_lobyte_first ( hdr.iOffsetSignature ) ;
	hdr.bSendTime          = int_from_lobyte_first ( hdr.bSendTime ) ;
	hdr.bSendResize        = int_from_lobyte_first ( hdr.bSendResize ) ;
	hdr.iRectLeft          = int_from_lobyte_first ( hdr.iRectLeft ) ;
	hdr.iRectTop           = int_from_lobyte_first ( hdr.iRectTop ) ;
	hdr.iRectRight         = int_from_lobyte_first ( hdr.iRectRight ) ;
	hdr.iRectBottom        = int_from_lobyte_first ( hdr.iRectBottom ) ;
	hdr.bSendCenter        = int_from_lobyte_first ( hdr.bSendCenter ) ;
	hdr.iOffsetFrom        = int_from_lobyte_first ( hdr.iOffsetFrom ) ;
	hdr.iOffsetTo          = int_from_lobyte_first ( hdr.iOffsetTo ) ;

	iOffsetMin = sizeof(struct OboegakiDataStruct) + strlen ( OBOEGAKI_STRUCT_STRING ) + 1 ;
	iOffsetMax = hdr.iStructSize ;

	if ( hdr.iStructSize < sizeof(struct OboegakiDataStruct)
	  || hdr.iStructVersion != OBOEGAKI_STRUCT_VERSION
	  || hdr.iOffsetData >= iOffsetMin  &&  hdr.iOffsetData >= iOffsetMax
	  || hdr.iOffsetFont >= iOffsetMin  &&  hdr.iOffsetFont >= iOffsetMax
	  || hdr.iOffsetFrom >= iOffsetMin  &&  hdr.iOffsetFrom >= iOffsetMax ) {
		xmemo_ErrorLog ( "xmemo_RecvAcceptedTcp_Oboegaki", "Discard", NULL ) ;
		return ;
	}

	pMemo = xmemo_CreateMemo ( ) ;
	if ( pMemo == NULL ) return ;

	pMemo->x = hdr.iRectLeft ;
	pMemo->y = hdr.iRectTop ;
	pMemo->w = hdr.iRectRight - hdr.iRectLeft + 1 ;
	pMemo->h = hdr.iRectBottom - hdr.iRectTop + 1 ;

	if      ( pMemo->w < MIN_DEFAULT_MEMO_WIDTH ) pMemo->w = iDefaultMemoWidth ;
	else if ( pMemo->w > MAX_DEFAULT_MEMO_WIDTH ) pMemo->w = MAX_DEFAULT_MEMO_WIDTH ;
	if      ( pMemo->h < MIN_DEFAULT_MEMO_HEIGHT ) pMemo->h = iDefaultMemoHeight ;
	else if ( pMemo->h > MAX_DEFAULT_MEMO_HEIGHT ) pMemo->h = MAX_DEFAULT_MEMO_HEIGHT ;
	if ( pMemo->x + pMemo->w > iScreenWidth ) pMemo->x = iScreenWidth - pMemo->w ;
	if ( pMemo->x < 0 ) pMemo->x = 0 ;
	if ( pMemo->y + pMemo->h > iScreenHeight ) pMemo->y = iScreenHeight - pMemo->h ;
	if ( pMemo->y < 0 ) pMemo->y = 0 ;

	pMemo->iBackColor[0] = ( hdr.iBackColor & 0x000000FF ) << 8 ;
	pMemo->iBackColor[1] = ( hdr.iBackColor & 0x0000FF00 )      ;
	pMemo->iBackColor[2] = ( hdr.iBackColor & 0x00FF0000 ) >> 8 ;
	pMemo->iFontColor[0] = ( hdr.iForeColor & 0x000000FF ) << 8 ;
	pMemo->iFontColor[1] = ( hdr.iForeColor & 0x0000FF00 )      ;
	pMemo->iFontColor[2] = ( hdr.iForeColor & 0x00FF0000 ) >> 8 ;
	if ( iOffsetMin <= hdr.iOffsetData  &&  hdr.iOffsetData < iOffsetMax ) {
		iMsgLen = xmemo_DecodeMessage_Oboegaki ( sMessage, &(data[sizeof(int)+hdr.iOffsetData]), MAXLEN_MESSAGE ) ;
		if ( iMsgLen > 0 ) {
			pMemo->pText = XtMalloc ( iMsgLen + 1 ) ;
			if ( pMemo->pText == NULL ) {
				xmemo_ErrorLog ( "xmemo_RecvAcceptedTcp_Oboegaki", "XtMalloc", NULL ) ;
				xmemo_FreeMemo ( pMemo ) ;
				return ;
			}
			strcpy ( pMemo->pText, sMessage ) ;
		}
	}
	if ( iOffsetMin <= hdr.iOffsetFrom  &&  hdr.iOffsetFrom < iOffsetMax ) {
		iMsgLen = xmemo_DecodeMessage_Oboegaki ( sMessage, &(data[sizeof(int)+hdr.iOffsetFrom]), MAXLEN_NICKNAME ) ;
		if ( iMsgLen > 0 ) {
			pMemo->pNick = XtMalloc ( iMsgLen + 1 ) ;
			if ( pMemo->pNick == NULL ) {
				xmemo_ErrorLog ( "xmemo_RecvAcceptedTcp_Oboegaki", "XtMalloc", NULL ) ;
				xmemo_FreeMemo ( pMemo ) ;
				return ;
			}
			strcpy ( pMemo->pNick, sMessage ) ;
		}
	}

	xmemo_InsertMemo ( pMemo ) ;

	/* Start timer */

	time ( &(pSocket->tTimeout) ) ;
	pSocket->tTimeout += XMEMO_OBOEGAKI_TIMEOUT ;
	xmemo_StartTimer ( pSocket->tTimeout ) ;

	return ;

}

/******************************************************************************/
/*  xmemo_AcceptTcp_Oboegaki                                                  */
/******************************************************************************/

void xmemo_AcceptTcp_Oboegaki ( zSocketType *pSocket,
                                struct sockaddr_in *addr, int addrlen, int iClientSock )
{

zSocketType	*pClientSocket ;
int 	rc ;

	if ( bMemoTraceLog ) {
		char	sLog [ 32 ] ;
		sprintf ( sLog, "%c://%s:%u",
			pSocket->cProtocol,
			ipaddress_long_to_string ( ntohl ( addr->sin_addr.s_addr ) ),
			(unsigned short) ntohs ( addr->sin_port ) ) ;
		xmemo_TraceLog ( "xmemo_AcceptTcp_Oboegaki", sLog ) ;
	}

	/* Send a connect message to the peer */

	rc = xmemo_SendAcceptedTcp_Oboegaki_Hello ( iClientSock ) ;
	if ( rc == RC_ERROR ) {
		close ( iClientSock ) ;
		return ;
	}

	/* Allocate memory */

	pClientSocket = (zSocketType*) XtMalloc ( sizeof(zSocketType) ) ;
	if ( pClientSocket == NULL ) {
		xmemo_ErrorLog ( "xmemo_AcceptTcp_Oboegaki", "XtMalloc", NULL ) ;
		close ( iClientSock ) ;
		return ;
	}

	pClientSocket->cSockType = XMEMO_SOCKET_TCP_ACCEPT ;
	pClientSocket->cProtocol = pSocket->cProtocol ;
	pClientSocket->iSock = iClientSock ;
	pClientSocket->iAddr = ntohl ( addr->sin_addr.s_addr ) ;
	pClientSocket->iPort = ntohs ( addr->sin_port ) ;
	pClientSocket->tTimeout = 0 ;
	pClientSocket->iCallbackReadId   = 0 ;
	pClientSocket->iCallbackWriteId  = 0 ;
	pClientSocket->iCallbackExceptId = 0 ;

	/* Insert into chained list */

	pClientSocket->pChain = xmemo_InsertIntoChain ( &zSocketAnchor, pClientSocket ) ;
	if ( pClientSocket->pChain == NULL ) {
		xmemo_ErrorLog ( "xmemo_AcceptTcp_Oboegaki", "xmemo_InsertIntoChain", NULL ) ;
		close ( pClientSocket->iSock ) ;
		XtFree ( (void*) pClientSocket ) ;
		return ;
	}

	/* Add callback */

	pClientSocket->iCallbackReadId = XtAppAddInput ( xApp, pClientSocket->iSock,
	                                                 (caddr_t) XtInputReadMask,
	                                                 (XtInputCallbackProc) xmemo_RecvAcceptedTcp_Oboegaki,
	                                                 (caddr_t) pClientSocket ) ;

	/* Start timer */

	time ( &(pClientSocket->tTimeout) ) ;
	pClientSocket->tTimeout += XMEMO_OBOEGAKI_TIMEOUT ;
	xmemo_StartTimer ( pClientSocket->tTimeout ) ;

	return ;

}

/******************************************************************************/
/*  xmemo_NetNotify_Oboegaki                                                  */
/******************************************************************************/

int xmemo_NetNotify_Oboegaki ( int iSock, long iAddr, short iPort, int iNotify )
{

int 	iOboegakiAbsence ;
struct	OboegakiEntryStruct	zOboegakiHeader ;
char	data [ 50 + MAXLEN_GROUPNAME + MAXLEN_NICKNAME + MAXLEN_USERNAME + MAXLEN_HOSTNAME ] ;
int 	iOffset ;
int 	rc ;
struct	sockaddr_in	addr ;

	if ( bMemoTraceLog ) {
		char	sLog [ 32 ] ;
		sprintf ( sLog, "%c://%s:%u",
			XMEMO_PROTOCOL_OBOEGAKI,
			ipaddress_long_to_string ( iAddr ),
			(unsigned short) iPort ) ;
		xmemo_TraceLog ( "xmemo_NetNotify_Oboegaki", sLog ) ;
	}

	iOffset = sizeof(struct OboegakiEntryStruct) ;

	strcpy ( zOboegakiHeader.sPacketString, "OboeBct" ) ;

	zOboegakiHeader.iPluginVersion = int_to_lobyte_first ( 0x00000100 ) ;

	if ( bUserAbsence ) iOboegakiAbsence = 1 ;
	else                iOboegakiAbsence = 0 ;

	switch ( iNotify ) {

	case XMEMO_NET_NOTIFY_ENTER :

		zOboegakiHeader.iNotify = int_to_lobyte_first ( OBOEGAKI_NOTIFY_ENTER ) ;
		zOboegakiHeader.bAbsence = int_to_lobyte_first ( iOboegakiAbsence ) ;
		break;

	case XMEMO_NET_NOTIFY_REPLY :

		zOboegakiHeader.iNotify = int_to_lobyte_first ( OBOEGAKI_NOTIFY_REPLY ) ;
		zOboegakiHeader.bAbsence = int_to_lobyte_first ( iOboegakiAbsence ) ;
		break;

	case XMEMO_NET_NOTIFY_LEAVE :

		zOboegakiHeader.iNotify = int_to_lobyte_first ( OBOEGAKI_NOTIFY_LEAVE ) ;
		zOboegakiHeader.bAbsence = int_to_lobyte_first ( iOboegakiAbsence ) ;
		break;

	case XMEMO_NET_NOTIFY_CHANGE :

		zOboegakiHeader.iNotify = int_to_lobyte_first ( OBOEGAKI_NOTIFY_REPLY ) ;
		zOboegakiHeader.bAbsence = int_to_lobyte_first ( iOboegakiAbsence ) ;
		break;

	case XMEMO_NET_NOTIFY_PRESENCE :

		zOboegakiHeader.iNotify = int_to_lobyte_first ( OBOEGAKI_NOTIFY_REPLY ) ;
		zOboegakiHeader.bAbsence = int_to_lobyte_first ( 0 ) ;
		break;

	case XMEMO_NET_NOTIFY_ABSENCE :

		zOboegakiHeader.iNotify = int_to_lobyte_first ( OBOEGAKI_NOTIFY_REPLY ) ;
		zOboegakiHeader.bAbsence = int_to_lobyte_first ( 1 ) ;
		break;

	}

	zOboegakiHeader.iOffsetNickname = int_to_lobyte_first ( iOffset ) ;
	xmemo_EncodeMessage_Oboegaki ( &data[iOffset], sMemoNickname, MAXLEN_NICKNAME ) ;
	iOffset += strlen ( &data[iOffset] ) + 1 ;

	zOboegakiHeader.iOffsetGroupname = int_to_lobyte_first ( iOffset ) ;
	xmemo_EncodeMessage_Oboegaki ( &data[iOffset], sMemoGroupname, MAXLEN_GROUPNAME ) ;
	iOffset += strlen ( &data[iOffset] ) + 1 ;

	zOboegakiHeader.iOffsetUsername = int_to_lobyte_first ( iOffset ) ;
	strcpy ( &data[iOffset], sMemoUsername ) ;
	iOffset += strlen ( sMemoUsername ) + 1 ;

	zOboegakiHeader.iOffsetHostname = int_to_lobyte_first ( iOffset ) ;
	strcpy ( &data[iOffset], sMemoHostname ) ;
	iOffset += strlen ( sMemoHostname ) + 1 ;

	zOboegakiHeader.iPacketSize = int_to_lobyte_first ( iOffset ) ;

	memcpy ( data, (char*)&zOboegakiHeader, sizeof(struct OboegakiEntryStruct) ) ;

	memset ( &addr, 0, sizeof(addr) ) ;
	addr.sin_family      = AF_INET ;
	addr.sin_port        = htons ( iPort ) ;
	addr.sin_addr.s_addr = htonl ( iAddr ) ;

	rc = sendto ( iSock, data, iOffset, 0, (struct sockaddr *)&addr, sizeof(addr) ) ;
	if ( rc >= 0 ) {
		rc = RC_SUCCESS ;
	} else {
		char	sLog [ 40 ] ;
		sprintf ( sLog, "%s:%d", ipaddress_long_to_string ( iAddr ), iPort ) ;
		xmemo_ErrorLog ( "xmemo_NetNotify_Oboegaki", "sendto", sLog ) ;
		rc = RC_ERROR ;
	}
	if ( bMemoTracePacket ) xmemo_TracePacket ( data, iOffset ) ;

	return rc ;

}
